 /*
 *Driver for the Analog Devices ADV7619 IC

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

 
#include <linux/kernel.h>
#include <linux/string.h>
#include "cxt_mgr.h"
#include "adv7619.h"
#include "infoframe.h"
#include "adv7619_reg.h"
#include "mem_model.h"
#include "i2c_model.h"
#include "sys.h"
#include "trace.h"
#include "debug.h"
#include "task_model.h"

#define AUDIO_32K		320000
#define AUDIO_44K		441000
#define AUDIO_48K		480000
#define AUDIO_88K		882000
#define AUDIO_96K		960000
#define AUDIO_176K		1764000
#define AUDIO_192K		1920000
#define AUDIO_TOLERANCE	50000

typedef enum
{
    IO_MAP=0,
    CP_MAP,
    HDMI_MAP,
    REPEATER_MAP,
    EDID_MAP,
    INFOFRAME_MAP,
    CEC_MAP,
    DPLL_MAP,
    ADV7619_MAP_COUNT,
}adv7619_map_e;

#if 1 
typedef struct
{
    BOOL_T  is_available;
    BOOL_T  interlaced;
    U32_T   width;
    U32_T   height;
    U32_T   framerate;
}unlock_enhance_frame_t;
#endif

typedef struct
{
    i2c_model_bus_handle_t bus_handle;
    U16_T addr;
    U8_T map_addr[ADV7619_MAP_COUNT];
    trace_attr_handle_t attr_handles[4];
    U8_T id;
    adv7619_callback_t callback_func;
    adv7619_callback_t_for_get_cardid call_fun_get_cardid;
    void *cb_cxt;
    void *cb_cxt_for_getid;
    task_handle_t task;
    task_model_handle_t taskmgr;
    adv7619_frameinfo_t frameinfo; 
    enum adv7619_audio_sample rates;
    int board_id;
    adv7619_event_e lock;
    U8_T avi_infoframe[INFOFRAME_MAXSIZE];
    U8_T vendor_infoframe[INFOFRAME_MAXSIZE];
    U8_T audio_infoframe[INFOFRAME_MAXSIZE];
    int desired_timing;
    int desired_framerate;
    int detected_audiorate;
#if 1 
    unlock_enhance_frame_t before_frame;
#endif
#if 1
    BOOL_T skip_unlock_enhance;
#endif
}adv7619_cxt_t;


typedef struct
{
    adv7619_map_e domain;
    U8_T reg;
    U8_T value;        
}regtbl_item_t;

typedef struct 
{
    U8_T alt_gamma; // 0: no convert 1: convert YUV601 <=> YUV709
    U8_T op_656_range; // 0: full, 1: limited
    U8_T rgb_out; // 0: YPbPr out 1: RGB out
    U8_T alt_data_sat; // 0:enable saturator according op_656_range 1:reverse op_656_range decision
}adv7619_out_color_ctrl_t;

typedef enum
{
    VSYNC_POS_MASK=(0x1<<0),
    HSYNC_POS_MASK=(0x1<<1),
}polarity_mask_e;

typedef struct
{
    polarity_mask_e  polarities;
    U64_T   pixelclock;
    U32_T   hfrontporch;
    U32_T   hsync;
    U32_T   hbackporch;
    U32_T   vfrontporch;
    U32_T   vsync;
    U32_T   vbackporch;
    U32_T   il_vfrontporch;
    U32_T   il_vsync;
    U32_T   il_vbackporch;
}detail_timing_t;

typedef struct
{
    U32_T   width;
    U32_T   height;
    BOOL_T  interlaced;
    U32_T   framerate;
    detail_timing_t detail;
}video_timing_t;

static int adv7619_wr_reg(adv7619_cxt_t *adv7619_cxt, U8_T reg, U8_T val);
static U8_T adv7619_rd_reg(adv7619_cxt_t *adv7619_cxt, U8_T reg);
static int adv7619_wr_map_reg(adv7619_cxt_t *adv7619_cxt,adv7619_map_e map, U8_T reg, U8_T val);
static U8_T adv7619_rd_map_reg(adv7619_cxt_t *adv7619_cxt,adv7619_map_e map, U8_T reg);
static int io_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T io_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int io_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val);
static int cp_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T cp_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int cp_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val);
static int cec_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T cec_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int dpll_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T dpll_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int edid_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T edid_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int hdmi_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T hdmi_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int hdmi_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val);
static int infoframe_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static U8_T infoframe_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static int repeater_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val);
static int repeater_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val);
static U8_T repeater_read(adv7619_cxt_t *adv7619_cxt,U8_T reg);
static void adv7619_reset(adv7619_cxt_t *adv7619_cxt);
static int adv7619_core_init(adv7619_cxt_t *adv7619_cxt);
static int adv7619_internal_edid(adv7619_cxt_t *adv7619_cxt);
static void adv7619_callback(adv7619_cxt_t *adv7619_cxt,adv7619_event_e event);
static void adv7619_enable_protection(adv7619_cxt_t *adv7619_cxt, BOOL_T enable);

static U8_T adv7619_dev_count=0;
static const unsigned char adv7619_edid[256];
static BOOL_T hdmi_encrpt=false;
U32_T g_card_id = 0;

static void adv7619_reset(adv7619_cxt_t *adv7619_cxt)
{
    io_write(adv7619_cxt, ADV7619_IO_REG_IO_REG_FF, ADV7619_IO_BIT_MAIN_RESET);   /* reset */
}

static regtbl_item_t regtbl_OperatingMode[]=
{
 //   { IO_MAP, 0x0c, 0x42 },
    { IO_MAP, 0x01, 0x06 },
    { IO_MAP, 0x00, 0x02 },
    { IO_MAP, 0x15, 0xa0 },

    //RxPll
    { HDMI_MAP, 0x50, 0xC0 },     /*PLL Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x57, 0xA3 },  //0222   /*PLL Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x58, 0x07 },  //0222   /*PLL Setting Synced with RSD Oct 2011*/

    // RxEq
    { HDMI_MAP, 0x5A, 0x80 },     /*ADI Recommended Eq Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x84, 0x03 },   /*ADI Recommended Eq Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x85, 0x10 },     /*ADI Recommended Eq Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x86, 0x9B },     /*ADI Recommended Eq Setting Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x9B, 0x03 },     /*ADI Recommended Eq Setting Synced with RSD Oct 2011*/

    //RxHdmiAdiRec
    { HDMI_MAP, 0x3E, 0x69 },     /*ADI Recommended HDMI Setting Synced with RSD Oct 2011*/
    { HDMI_MAP, 0x3F, 0x46 },     /*ADI Recommended HDMI Setting Synced with RSD Oct 2011*/
    { HDMI_MAP, 0x4E, 0xFE },     /*ADI Recommended HDMI Setting *Updated RSD Oct 2011*/
    { HDMI_MAP, 0x4F, 0x08 },     /*ADI Recommended HDMI Setting Updated RSD Oct 2011*/
    { HDMI_MAP, 0x93, 0x03 },     /*ADI Recommended HDMI Setting: PRC Updated RSD Oct 2011*/
    { HDMI_MAP, 0xC0, 0x03 },     /*ADI Recommended HDMI Setting: Power Setting  Synced with RSD Oct 2011*/
    //{ HDMI_MAP, 0x83, 0xFC },     /* ADI recommended write    */
    //{ HDMI_MAP, 0xBA, 0xB2 },      /* ADI recommended write    */ /*to review: this set DLL2_bypass to 1*/
    { HDMI_MAP, 0x10, 0xA5 },     /* Turn ON HDCP mute flag bit 7  Synced with RSD Oct 2011*/
    { HDMI_MAP, 0x45, 0x04 },     /* Set Burst Err Filter, prevent PS3 audio noise Synced with RSD Oct 2011*/
    { HDMI_MAP, 0x4C, 0x44 },     /* Vertical Filter selection: Added 18 October, due to Vsync lock instability resulting in 2 resolution detections*/
    { IO_MAP, 0x19, 0x96 },       /* DLL setting- can be reset in application layer by end user as per hw */
    { IO_MAP, 0x14, 0x7f }, //0222
    { IO_MAP, 0x33, 0x40 },       /* DLL setting- can be reset in application layer by end user as per hw */
    { CP_MAP, 0xC8, 0x08 },       /* Set digital fine clamp setting for HDMI Mode*/
    { CP_MAP, 0x6C, 0x00 },       /* Disable Auto Clamping-Synced with RSD Oct 2011*/
    { HDMI_MAP, 0x50, 0x00 },     /* Adjust Gear Wait Time to Minimum HDCP @ 4k2k Interconnectivity*/
    { HDMI_MAP, 0x01, 0x01 },     /* Enable termination auto */
#if 1    
    { IO_MAP,   0xDD, 0x00 },	  //need re-set this value in this mode
    //{ IO_MAP,   0x02, 0xF5 },   //(Fixed color issue for HD) Set OP_656_RANGE to 1, refer to ADV7619 evboard.
    { DPLL_MAP, 0xC3, 0x00 },	  //need re-set this value in this mode
    { DPLL_MAP, 0xCF, 0x00 }, 	  //need re-set this value in this mode
                                  // ADI recommend setting for 720p,1080i,1080p color depths 24 and 36
    { CP_MAP,   0x6C, 0x00 },           
    { HDMI_MAP, 0xC0, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x08},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3D, 0x10},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    #if 0
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    #endif
    { HDMI_MAP, 0x57, 0xE5}, //0222
    { HDMI_MAP, 0x58, 0x1C}, //0222
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x84, 0x03},  
    
    { HDMI_MAP, 0x8F, 0x10}, //0222    
    { HDMI_MAP, 0x90, 0x63}, //0222              
                                    
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x14}, //0225
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x92, 0x08}, //0222
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
#endif    
};

#if 0
static regtbl_item_t Setting_SD[] = {
    { IO_MAP, 0xDD, 0x00 },	        // need re-set this value in this mode
    { IO_MAP, 0x02, 0xF4 },  		// Set OP_656_RANGE to 1, refer to ADV7619 evboard.
    { DPLL_MAP, 0xC3, 0x00 },		// need re-set this value in this mode
    { DPLL_MAP, 0xCF, 0x00 }, 		// need re-set this value in this mode
                                    // ADI recommend setting for 480i,576i,480p,576p color depths 24 and 36
    { CP_MAP,   0x6C, 0x00 },             
    { HDMI_MAP, 0xC0, 0x03 },
    { HDMI_MAP, 0x03, 0x98 },
    { HDMI_MAP, 0x10, 0xA5 },
    { HDMI_MAP, 0x1B, 0x08 },
    { HDMI_MAP, 0x45, 0x04 },
    { HDMI_MAP, 0x97, 0xC0 },
    { HDMI_MAP, 0x3D, 0x10 },
    { HDMI_MAP, 0x3E, 0x69 },
    { HDMI_MAP, 0x3F, 0x46 },
    { HDMI_MAP, 0x4E, 0xFE },
    { HDMI_MAP, 0x4F, 0x08 },
    { HDMI_MAP, 0x50, 0x00 },
    { HDMI_MAP, 0x57, 0xA3 },
    { HDMI_MAP, 0x58, 0x07 },
    { HDMI_MAP, 0x6f, 0x08 },
    { HDMI_MAP, 0x84, 0x03 },           
                                        
	
    // the followng is for EQ
    { HDMI_MAP, 0x85, 0x11 },
    { HDMI_MAP, 0x86, 0x9B },
    { HDMI_MAP, 0x89, 0x03 },
    { HDMI_MAP, 0x9B, 0x03 },
    { HDMI_MAP, 0x93, 0x03 },
    { HDMI_MAP, 0x5A, 0x80 },
    { HDMI_MAP, 0x9C, 0x80 },
    { HDMI_MAP, 0x9C, 0xC0 },
    { HDMI_MAP, 0x9C, 0x00 } 
};
#endif

static regtbl_item_t Setting_SD[] = {
	#if 0
    { HDMI_MAP, 0xC0, 0x03},
	//{ IO_MAP, 0x00, 0x19},
	{ IO_MAP, 0x01, 0x06},  
	////{ IO_MAP, 0x02, 0xF7}, /////
	{ IO_MAP, 0x02, 0xF4}, 
	{ IO_MAP, 0x03, 0x40}, 
	{ IO_MAP, 0x05, 0x28}, 
	{ IO_MAP, 0x06, 0xA0}, 
    { IO_MAP, 0x0C, 0x42}, 
	{ IO_MAP, 0x15, 0x80},
	{ IO_MAP, 0x19, 0x83},
	{ IO_MAP, 0x33, 0x40},  
	//{ IO_MAP, 0xDD, 0x00},  
	//{ IO_MAP, 0xE7, 0x04},
	#else
	{ IO_MAP, 0xDD, 0x00 },	      
    { IO_MAP, 0x02, 0xF4 },  
	#endif 
	{ CP_MAP,   0xBA, 0x01 }, 
	{ CP_MAP,   0x6C, 0x00 }, 
	{ DPLL_MAP, 0xB5, 0x01},
	//{ DPLL_MAP, 0xC3, 0x80},
    //{ DPLL_MAP, 0xCF, 0x03}, 
    //{ DPLL_MAP, 0xDB, 0x80}, 
    { HDMI_MAP, 0xC0, 0x03},
    { HDMI_MAP, 0x00, 0x08},
    { HDMI_MAP, 0x02, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x08},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3D, 0x10},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x83, 0xFC},
    { HDMI_MAP, 0x84, 0x03},
    
   
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x11},
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
};

#if 0
//720P, 1080I AND 1080P RESOLUTIONS UP TO A 36-BIT COLOR DEPTH

static regtbl_item_t Setting_HD[] = {
    { IO_MAP,   0xDD, 0x00 }, //need re-set this value in this mode
    { IO_MAP,   0x02, 0xF4 }, //Set OP_656_RANGE to 1, refer to ADV7619 evboard.
    { DPLL_MAP, 0xC3, 0x00 }, //need re-set this value in this mode
    { DPLL_MAP, 0xCF, 0x00 }, //need re-set this value in this mode
                              // ADI recommend setting for 720p,1080i,1080p color depths 24 and 36
    { CP_MAP,   0x6C, 0x00 },           
    { HDMI_MAP, 0xC0, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x08},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3D, 0x10},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x84, 0x03},            
                                    
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x10},
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
};
#endif

static regtbl_item_t Setting_HD[] = {
	{ HDMI_MAP, 0xC0, 0x03},
	#if 0
	//{ IO_MAP, 0x00, 0x19},
	{ IO_MAP, 0x01, 0x06},  
	//{ IO_MAP, 0x02, 0xF7}, //////
	{ IO_MAP, 0x02, 0xF4},
	{ IO_MAP, 0x03, 0x40}, 
	{ IO_MAP, 0x05, 0x28}, 
	{ IO_MAP, 0x06, 0xA6}, 
    { IO_MAP, 0x0C, 0x42}, 
	{ IO_MAP, 0x15, 0x80},
	{ IO_MAP, 0x19, 0x83},
	{ IO_MAP, 0x33, 0x40},  
	//{ IO_MAP, 0xDD, 0x00},  
	//{ IO_MAP, 0xE7, 0x04}, 
	#else
	{ IO_MAP,   0xDD, 0x00 }, 
    { IO_MAP,   0x02, 0xF4 },
	#endif
	{ CP_MAP,   0xBA, 0x01 }, 
	{ CP_MAP,   0x6C, 0x00 }, 
	{ DPLL_MAP, 0xB5, 0x01},
	//{ DPLL_MAP, 0xC3, 0x80},
    //{ DPLL_MAP, 0xCF, 0x03}, 
    //{ DPLL_MAP, 0xDB, 0x80}, 
    { HDMI_MAP, 0xC0, 0x03},
    //{ HDMI_MAP, 0x00, 0x08},//////////
    { HDMI_MAP, 0x02, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x08},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x84, 0x03},
                            
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x10},
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
};

//	4K2K RESOLUTION WITH A 24 - BIT COLOR DEPTH
#if 0
static regtbl_item_t Setting_4k2k[] = {
    { IO_MAP, 0x02, 0xF4}, //(Recovery to 0) Set OP_656_RANGE to 0.
                           // ADI recommend setting for 4K2K color depths 24 
    { IO_MAP, 0xDD, 0xA0}, //  LLC Half frequency
    { DPLL_MAP, 0xC3, 0x80},
    { DPLL_MAP, 0xCF, 0x03}, // 4c->30 doc error?
    { HDMI_MAP, 0xC0, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x00},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x84, 0x03},
                            
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x10},
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
};
#endif

static regtbl_item_t Setting_4k2k[] = {
	
	{ HDMI_MAP, 0xC0, 0x03},
	#if 0
	{ IO_MAP, 0x00, 0x19},
	{ IO_MAP, 0x01, 0x05},  
	{ IO_MAP, 0x02, 0xF2}, 
	{ IO_MAP, 0x03, 0x54}, 
	{ IO_MAP, 0x05, 0x28}, 
	{ IO_MAP, 0x06, 0xA0}, 
    { IO_MAP, 0x0C, 0x42}, 
	{ IO_MAP, 0x15, 0x80},
	{ IO_MAP, 0x19, 0x83},
	{ IO_MAP, 0x33, 0x40},  
	{ IO_MAP, 0xDD, 0x00},  
	{ IO_MAP, 0xE7, 0x04},
	#else
	{ IO_MAP, 0x02, 0xF4},                          
    { IO_MAP, 0xDD, 0xA0}, 
	#endif 
	{ DPLL_MAP, 0xB5, 0x01},
	{ DPLL_MAP, 0xC3, 0x80},
    { DPLL_MAP, 0xCF, 0x03}, 
    { DPLL_MAP, 0xDB, 0x80}, 
    { HDMI_MAP, 0xC0, 0x03},
    { HDMI_MAP, 0x00, 0x08},
    { HDMI_MAP, 0x02, 0x03},
    { HDMI_MAP, 0x03, 0x98},
    { HDMI_MAP, 0x10, 0xA5},
    { HDMI_MAP, 0x1B, 0x00},
    { HDMI_MAP, 0x45, 0x04},
    { HDMI_MAP, 0x97, 0xC0},
    { HDMI_MAP, 0x3E, 0x69},
    { HDMI_MAP, 0x3F, 0x46},
    { HDMI_MAP, 0x4E, 0xFE},
    { HDMI_MAP, 0x4F, 0x08},
    { HDMI_MAP, 0x50, 0x00},
    #if 0
    { HDMI_MAP, 0x57, 0xA3},
    { HDMI_MAP, 0x58, 0x07},
    #endif
    { HDMI_MAP, 0x57, 0xE5}, //0222
    { HDMI_MAP, 0x58, 0x1C}, //0222
    { HDMI_MAP, 0x6F, 0x08},
    { HDMI_MAP, 0x84, 0x03},
    { HDMI_MAP, 0x8F, 0x10}, //0222
    { HDMI_MAP, 0x90, 0x63}, //0222
                            
                            
    //  the followng is for EQ
    { HDMI_MAP, 0x85, 0x14}, //0225
    { HDMI_MAP, 0x86, 0x9B},
    { HDMI_MAP, 0x89, 0x03},
    { HDMI_MAP, 0x9B, 0x03},
    { HDMI_MAP, 0x92, 0x08}, //0222
    { HDMI_MAP, 0x93, 0x03},
    { HDMI_MAP, 0x5A, 0x80},
    { HDMI_MAP, 0x9C, 0x80},
    { HDMI_MAP, 0x9C, 0xC0},
    { HDMI_MAP, 0x9C, 0x00}
};

#define EDID_BYTES 256

// 7619 edid buffer, 1920x1080
static const unsigned char adv7619_edid_cm311h[EDID_BYTES]= 
{
	0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
	0x06, 0xd8, 0x28, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x31, 0x1a, 0x01, 0x03, 0x80, 0x50, 0x2d, 0x78,
	0x22, 0x76, 0x90, 0xa8, 0x54, 0x4d, 0x9f, 0x25,
	0x0e, 0x50, 0x54, 0xff, 0xff, 0x80, 0x95, 0x00,
	0x81, 0xc0, 0x81, 0x40, 0x81, 0x80, 0xa9, 0x40,
	0x90, 0x40, 0xb3, 0x00, 0x81, 0x00, 0x02, 0x3A,
	0x80, 0x18, 0x71, 0x38, 0x2d, 0x40, 0x58, 0x2c,
	0x45, 0x00, 0xc4, 0x8e, 0x21, 0x00, 0x00, 0x1e,
	0x84, 0x21, 0x56, 0xaa, 0x51, 0x00, 0x1e, 0x30,
	0x46, 0x8f, 0x33, 0x00, 0xc4, 0x8e, 0x21, 0x00,
	0x00, 0x1c, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x41,
	0x56, 0x54, 0x20, 0x43, 0x4d, 0x33, 0x31, 0x31,
	0x48, 0x0a, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfd,
	0x00, 0x17, 0x4c, 0x1e, 0x8c, 0x3c, 0x00, 0x0a,
	0x00, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0xed,
	0x02, 0x03, 0x1f, 0xf1, 0x4c, 0x10, 0x05, 0x04,
	0x03, 0x02, 0x20, 0x11, 0x01, 0x1f, 0x12, 0x13,
	0x14, 0x23, 0x09, 0x07, 0x07, 0x83, 0x01, 0x00,
	0x00, 0x65, 0x03, 0x0c, 0x00, 0x10, 0x00, 0x5d,
	0x21, 0x58, 0xa0, 0x51, 0x00, 0x1e, 0x30, 0x48,
	0x88, 0x3a, 0x00, 0xc4, 0x8e, 0x21, 0x00, 0x00,
	0x1c, 0x28, 0x3c, 0x80, 0xa0, 0x70, 0xb0, 0x23,
	0x40, 0x30, 0x20, 0x36, 0x00, 0xc4, 0x8e, 0x21,
	0x00, 0x00, 0x1a, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xed
};

// 7619 edid buffer, 4kx2k
static const unsigned char adv7619_edid_cn311h[EDID_BYTES]= 
{
	0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
      0x06, 0xd8, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x1a, 0x1b, 0x01, 0x03, 0x80, 0x50, 0x2d, 0x78,
      0x22, 0x76, 0x90, 0xa8, 0x54, 0x4d, 0x9f, 0x25,
      0x0e, 0x50, 0x54, 0xff, 0xff, 0x80, 0x95, 0x00,
      0xe1, 0x40, 0x81, 0x40, 0x81, 0x80, 0xa9, 0x40,
      0x90, 0x40, 0xb3, 0x00, 0x81, 0x00, 0x04, 0x74,
      0x00, 0x30, 0xf2, 0x70, 0x5a, 0x80, 0xb0, 0x58,
      0x8a, 0x00, 0x54, 0x4f, 0x21, 0x00, 0x00, 0x1e,
      0x02, 0x3a, 0x80, 0x18, 0x71, 0x38, 0x2d, 0x40,
      0x58, 0x2c, 0x45, 0x00, 0xc4, 0x8e, 0x21, 0x00,
      0x00, 0x1c, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x41,
      0x56, 0x54, 0x20, 0x43, 0x4e, 0x33, 0x31, 0x31,
      0x48, 0x0a, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfd,
      0x00, 0x17, 0x4c, 0x1e, 0x8c, 0x1e, 0x00, 0x0a,
      0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0x32,
      0x02, 0x03, 0x37, 0xf1, 0x54, 0x10, 0x05, 0x04,
      0x03, 0x02, 0x20, 0x11, 0x01, 0x1f, 0x12, 0x13,
      0x14, 0x5d, 0x5e, 0x5f, 0x5a, 0x59, 0x58, 0x57,
      0x56, 0x23, 0x09, 0x07, 0x07, 0x83, 0x01, 0x00,
      0x00, 0x6d, 0x03, 0x0c, 0x00, 0x10, 0x00, 0x00,
      0x3c, 0x20, 0x00, 0x60, 0x01, 0x02, 0x03, 0x67,
      0xd8, 0x5d, 0xc4, 0x01, 0x3c, 0xc0, 0x00, 0x5d,
      0x21, 0x58, 0xa0, 0x51, 0x00, 0x1e, 0x30, 0x48,
      0x88, 0x3a, 0x00, 0xc4, 0x8e, 0x21, 0x00, 0x00,
      0x1c, 0x28, 0x3c, 0x80, 0xa0, 0x70, 0xb0, 0x23,
      0x40, 0x30, 0x20, 0x36, 0x00, 0xc4, 0x8e, 0x21,
      0x00, 0x00, 0x1a, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xd4
};

static const unsigned char adv7619_edid_cn311h_4k[EDID_BYTES]= 
{
	0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
	0x06, 0xd8, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x1a, 0x1b, 0x01, 0x03, 0x80, 0x00, 0x00, 0x78,
	0x0a, 0x76, 0x90, 0xa8, 0x54, 0x4d, 0x9f, 0x25,
	0x0e, 0x50, 0x54, 0x20, 0x00, 0x00, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x04, 0x74,
	0x00, 0x30, 0xf2, 0x70, 0x5a, 0x80, 0xb0, 0x58,
	0x8a, 0x00, 0x54, 0x4f, 0x21, 0x00, 0x00, 0x1e,
	0x02, 0x3a, 0x80, 0x18, 0x71, 0x38, 0x2d, 0x40,
	0x58, 0x2c, 0x45, 0x00, 0xc4, 0x8e, 0x21, 0x00,
	0x00, 0x1c, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x41,
	0x56, 0x54, 0x20, 0x43, 0x4e, 0x33, 0x31, 0x31,
	0x48, 0x0a, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfd,
	0x00, 0x17, 0x4c, 0x1e, 0x8c, 0x1e, 0x00, 0x0a,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0x7a,
	
	0x02, 0x03, 0x23, 0xf0, 0x49, 0x5f, 0x5e, 0x10,
	0x1f, 0x04, 0x13, 0x02, 0x11, 0x01, 0x23, 0x09,
	0x07, 0x07, 0x83, 0x01, 0x00, 0x00, 0x6c, 0x03,
	0x0c, 0x00, 0x10, 0x00, 0x00, 0x3c, 0x20, 0x00,
	0x40, 0x01, 0x02, 0x01, 0x1d, 0x00, 0x72, 0x51,
	0xd0, 0x1e, 0x20, 0x6e, 0x28, 0x55, 0x00, 0xc4,
	0x8e, 0x21, 0x00, 0x00, 0x1e, 0x01, 0x1d, 0x00,
	0xbc, 0x52, 0xd0, 0x1e, 0x20, 0xb8, 0x28, 0x55,
	0x40, 0xc4, 0x8e, 0x21, 0x00, 0x00, 0x1e, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf5
};

static const unsigned char adv7619_edid_cn311h_1080p[EDID_BYTES]= 
{
	0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
	0x06, 0xd8, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x1a, 0x1b, 0x01, 0x03, 0x80, 0x00, 0x00, 0x78,
	0x0a, 0x76, 0x90, 0xa8, 0x54, 0x4d, 0x9f, 0x25,
	0x0e, 0x50, 0x54, 0x20, 0x00, 0x00, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x02, 0x3a,
	0x80, 0x18, 0x71, 0x38, 0x2d, 0x40, 0x58, 0x2c,
	0x45, 0x00, 0xc4, 0x8e, 0x21, 0x00, 0x00, 0x1e,
	0x02, 0x3a, 0x80, 0xd0, 0x72, 0x38, 0x2d, 0x40,
	0x10, 0x2c, 0x45, 0x80, 0xc4, 0x8e, 0x21, 0x00,
	0x00, 0x1e, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x41,
	0x56, 0x54, 0x20, 0x43, 0x4e, 0x33, 0x31, 0x31,
	0x48, 0x0a, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfd,
	0x00, 0x17, 0x4c, 0x1e, 0x8c, 0x0f, 0x00, 0x0a,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0xaa,
	
	0x02, 0x03, 0x1c, 0xf0, 0x47, 0x10, 0x1f, 0x04,
	0x13, 0x02, 0x11, 0x01, 0x23, 0x09, 0x07, 0x07,
	0x83, 0x01, 0x00, 0x00, 0x67, 0x03, 0x0c, 0x00,
	0x10, 0x00, 0x00, 0x1e, 0x01, 0x1d, 0x00, 0x72,
	0x51, 0xd0, 0x1e, 0x20, 0x6e, 0x28, 0x55, 0x00,
	0xc4, 0x8e, 0x21, 0x00, 0x00, 0x1e, 0x01, 0x1d,
	0x00, 0xbc, 0x52, 0xd0, 0x1e, 0x20, 0xb8, 0x28,
	0x55, 0x40, 0xc4, 0x8e, 0x21, 0x00, 0x00, 0x1e,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x41
};

void adv7619_get_brightness(handle_t adv7619_handle,int *BCHS_value)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	int value;
	
    value =cp_read(adv7619_cxt,0x3c);
 
	*BCHS_value = value;
	   
	if (*BCHS_value<=127)
        *BCHS_value+=128;
    else
		*BCHS_value-=128;	  
}

void adv7619_get_contrast(handle_t adv7619_handle,int *BCHS_value)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	int value;
	
    value =cp_read(adv7619_cxt,0x3a);
		   	  
	*BCHS_value = value;
}

void adv7619_get_hue(handle_t adv7619_handle,int *BCHS_value)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	int value;
	
    value =cp_read(adv7619_cxt,0x3d);
		   	  
	*BCHS_value = value;
}

void adv7619_get_saturation(handle_t adv7619_handle,int *BCHS_value)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	int value;
	
    value =cp_read(adv7619_cxt,0x3b);
		     
	*BCHS_value = value;
}

void adv7619_set_bchs(handle_t adv7619_handle,int *BCHS_value,int *bchs_select)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	int bchs_sel = *bchs_select;

	switch (bchs_sel)
	{
		case ADV7619_BCHS_BRIGHTNESS:
		    if (*BCHS_value<=127)
		        *BCHS_value+=128;
		    else
		        *BCHS_value-=128;
		  
		    cp_write(adv7619_cxt,0x3c,*BCHS_value);
		    break;
		    
		case ADV7619_BCHS_CONTRAST:
		    cp_write(adv7619_cxt,0x3a,*BCHS_value);
		    break;  
		
		case ADV7619_BCHS_HUE:
		    cp_write(adv7619_cxt,0x3d,*BCHS_value);
		    break;          
		
		case ADV7619_BCHS_SATURATION:
		    cp_write(adv7619_cxt,0x3b,*BCHS_value);
		    break;      
		    
	}
	cp_write(adv7619_cxt,0x3e,0x80); //enable
}
    
static void adv7619_set_out_color(adv7619_cxt_t *adv7619_cxt,adv7619_out_color_ctrl_t *out_color_ctrl)
{
    U8_T value=0;
    adv7619_out_color_ctrl_t *out_color_set = out_color_ctrl;
    
    if(out_color_set->alt_gamma)
         value |=ADV7619_IO_BIT_ALT_GAMMA; 
    if(out_color_set->op_656_range)
        value |= ADV7619_IO_BIT_OP_656_RANGE; 
    if(out_color_set->rgb_out)
        value |= ADV7619_IO_BIT_RGB_OUT; 
    if(out_color_set->alt_data_sat)
        value |= ADV7619_IO_BIT_ALT_DATA_SAT;
    
    io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_ALT_GAMMA|ADV7619_IO_BIT_OP_656_RANGE|ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_ALT_DATA_SAT,value);
    
}

static int adv7619_core_init(adv7619_cxt_t *adv7619_cxt)
{
    int i;
    regtbl_item_t *item;
    io_mask_write(adv7619_cxt, ADV7619_IO_REG_IO_REG_0C, (ADV7619_IO_BIT_POWER_DOWN|ADV7619_IO_BIT_CP_PWRDN|ADV7619_IO_BIT_PADS_PDN),0);   /* chip and cp core pad spowerup*/
    for(i=0,item=regtbl_OperatingMode;i<sizeof(regtbl_OperatingMode)/sizeof(regtbl_item_t);i++,item++)
    {
        adv7619_wr_map_reg(adv7619_cxt,item->domain,item->reg,item->value);
    }
    return 0;
}

static int adv7619_wr_map_reg(adv7619_cxt_t *adv7619_cxt,adv7619_map_e map, U8_T reg, U8_T val)
{
    U8_T wrbuf[2];
    int ret = 0;

    i2c_model_msg_t msg = 
    {
        .addr = adv7619_cxt->map_addr[map],
        .flags = 0,
        .buf = wrbuf,
        .len = 2
    };

    wrbuf[0] = reg;
    wrbuf[1] = val;

    ret = i2c_model_transfer(adv7619_cxt->bus_handle, &msg, 1);

    if(ret != 1) 
    {
        debug_msg("%s i2c_transfer failed %d\n", __func__, ret);
    }
    else 
    {
        return 0;
    }

    return -1;
}

static U8_T adv7619_rd_map_reg(adv7619_cxt_t *adv7619_cxt,adv7619_map_e map, U8_T reg)
{

    U8_T wrbuf = reg;
    U8_T rdbuf = 0;
    int ret = 0;

    i2c_model_msg_t msg[2] = 
    {
        {
            .addr = adv7619_cxt->map_addr[map],
            .flags = 0,
            .buf = &wrbuf,
            .len = 1
        }, {
            .addr = adv7619_cxt->map_addr[map],
            .flags = I2C_MODEL_READ,
            .buf = &rdbuf,
            .len = 1
        }
    };

    ret = i2c_model_transfer(adv7619_cxt->bus_handle, msg, 2);

    if(ret != 2) 
    {
        debug_msg("%s i2c_transfer failed %d\n", __func__, ret);
    }
    else 
    {
        return rdbuf;
    }

    return -1;
}

static int io_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val) 
{
    return adv7619_wr_map_reg(adv7619_cxt,IO_MAP,reg,val); 
}

static U8_T io_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,IO_MAP,reg);
}

static int io_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val)
{
    U8_T old_value,value;
    int ret=-1;
    if(adv7619_cxt)
    {
       old_value =io_read(adv7619_cxt,reg);
       value=old_value & (~mask);
       value |= (val & mask);
       ret=io_write(adv7619_cxt,reg,value);
    }
    return ret;
}

static int cp_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,CP_MAP,reg,val);
}

static U8_T cp_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,CP_MAP,reg);
}

static int cp_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val)
{
    U8_T old_value,value;
    int ret=-1;
    if(adv7619_cxt)
    {
       old_value =cp_read(adv7619_cxt,reg);
       value=old_value & (~mask);
       value|= val & mask;
       ret=cp_write(adv7619_cxt,reg,value);
    }
    return ret;
}


static int hdmi_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,HDMI_MAP,reg,val);
}

static U8_T hdmi_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,HDMI_MAP,reg);
}

static int hdmi_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val)
{
    U8_T old_value,value;
    int ret=-1;
    if(adv7619_cxt)
    {
       old_value =hdmi_read(adv7619_cxt,reg);
       value=old_value & (~mask);
       value|= val & mask;
       ret=hdmi_write(adv7619_cxt,reg,value);
    }
    return ret;
}

static int repeater_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val) 
{
    return adv7619_wr_map_reg(adv7619_cxt,REPEATER_MAP,reg,val);
}

static U8_T repeater_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)  
{
    return adv7619_rd_map_reg(adv7619_cxt,REPEATER_MAP,reg);
}

static int repeater_mask_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T mask,U8_T val)
{
    U8_T old_value,value;
    int ret=-1;
    if(adv7619_cxt)
    {
       old_value =repeater_read(adv7619_cxt,reg);
       value=old_value & (~mask);
       value|= val & mask;
       ret=repeater_write(adv7619_cxt,reg,value);
    }
    return ret;
}

static int edid_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,EDID_MAP,reg,val);
}

static U8_T edid_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,EDID_MAP,reg);
}

static __attribute__ ((unused)) int infoframe_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,INFOFRAME_MAP,reg,val);
}

static U8_T infoframe_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,INFOFRAME_MAP,reg);
}

static __attribute__ ((unused))int cec_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,CEC_MAP,reg,val);
}

static __attribute__ ((unused)) U8_T cec_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,CEC_MAP,reg);
}

static __attribute__ ((unused)) int dpll_write(adv7619_cxt_t *adv7619_cxt,U8_T reg,U8_T val)
{
    return adv7619_wr_map_reg(adv7619_cxt,DPLL_MAP,reg,val);
}

static __attribute__ ((unused)) U8_T dpll_read(adv7619_cxt_t *adv7619_cxt,U8_T reg)
{
    return adv7619_rd_map_reg(adv7619_cxt,DPLL_MAP,reg);
}

static int adv7619_wr_reg(adv7619_cxt_t *adv7619_cxt, U8_T reg, U8_T val)
{
    U8_T wrbuf[2];
    int ret = 0;

    i2c_model_msg_t msg = {
        .addr = adv7619_cxt->addr,
        .flags = 0,
        .buf = wrbuf,
        .len = 2
    };

    wrbuf[0] = reg;
    wrbuf[1] = val;

    ret = i2c_model_transfer(adv7619_cxt->bus_handle, &msg, 1);

    if(ret != 1) {
        debug_msg("%s i2c_transfer failed %d\n", __func__, ret);
    }
    else 
    {
        return 0;
    }

    return -1;
}

static U8_T adv7619_rd_reg(adv7619_cxt_t *adv7619_cxt, U8_T reg)
{

    U8_T wrbuf = reg;
    U8_T rdbuf = 0;
    int ret = 0;

    i2c_model_msg_t msg[2] = 
    {
        {
            .addr = adv7619_cxt->addr,
            .flags = 0,
            .buf = &wrbuf,
            .len = 1
        }, {
            .addr = adv7619_cxt->addr,
            .flags = I2C_MODEL_READ,
            .buf = &rdbuf,
            .len = 1
        }
    };

    ret = i2c_model_transfer(adv7619_cxt->bus_handle, msg, 2);

    if(ret != 2) 
    {
        debug_msg("%s i2c_transfer failed %d\n", __func__, ret);
    }
    else 
    {
        return rdbuf;
    }

    return -1;
}

static unsigned int adv7619io_show(void *cxt,  char *buf)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	int i;
	int c;
	int columns=16;

	c=0;
	c+=sys_sprintf(&buf[c],"    ");
	for(i=0;i<columns;i++)
		c+=sys_sprintf(&buf[c],"%02x ",i);
	c+=sys_sprintf(&buf[c],"\n");

	c+=sys_sprintf(&buf[c],"____");
	for(i=0;i<columns;i++)
			c+=sys_sprintf(&buf[c],"___");
	c+=sys_sprintf(&buf[c],"\n");

	for(i=0;i<0x100;i++)
	{
		if((i% columns)==0)
		{
			if(i>0 )
                            c+=sys_sprintf(&buf[c],"\n");
			c+=sys_sprintf(&buf[c],"%02x: ",i);
		}
		c+=sys_sprintf(&buf[c],"%02x ",io_read(adv7619_cxt,i));
	}
	c+=sys_sprintf(&buf[c],"\n");
	return c;
}

static unsigned int adv7619io_store(void *cxt, const char *buf, unsigned int count)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	unsigned reg;
	unsigned val;

	if(sys_sscanf(buf,"%x=%x", &reg,&val)==2)
	{
		debug_msg("set reg %x = %x\n",reg,val);
		io_write(adv7619_cxt,reg,val);
	}
    return count;
}

static unsigned int adv7619cp_show(void *cxt,  char *buf)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	int i;
	int c;
	int columns=16;

	c=0;
	c+=sys_sprintf(&buf[c],"    ");
	for(i=0;i<columns;i++)
		c+=sys_sprintf(&buf[c],"%02x ",i);
	c+=sys_sprintf(&buf[c],"\n");

	c+=sys_sprintf(&buf[c],"____");
	for(i=0;i<columns;i++)
			c+=sys_sprintf(&buf[c],"___");
	c+=sys_sprintf(&buf[c],"\n");

	for(i=0;i<0x100;i++)
	{
		if((i% columns)==0)
		{
			if(i>0 )
					c+=sys_sprintf(&buf[c],"\n");
			c+=sys_sprintf(&buf[c],"%02x: ",i);
		}
		c+=sys_sprintf(&buf[c],"%02x ",cp_read(adv7619_cxt,i));
	}
	c+=sys_sprintf(&buf[c],"\n");
  return c;
}

static unsigned int adv7619cp_store(void *cxt, const char *buf, unsigned int count)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	unsigned reg;
	unsigned val;

	if(sys_sscanf(buf,"%x=%x", &reg,&val)==2)
	{
		debug_msg("set reg %x = %x\n",reg,val);
		cp_write(adv7619_cxt,reg,val);
	}
    return count;
}

static unsigned int adv7619hdmi_show(void *cxt,  char *buf)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	int i;
	int c;
	int columns=16;

	c=0;
	c+=sys_sprintf(&buf[c],"    ");
	for(i=0;i<columns;i++)
		c+=sys_sprintf(&buf[c],"%02x ",i);
	c+=sys_sprintf(&buf[c],"\n");

	c+=sys_sprintf(&buf[c],"____");
	for(i=0;i<columns;i++)
			c+=sys_sprintf(&buf[c],"___");
	c+=sys_sprintf(&buf[c],"\n");

	for(i=0;i<0x100;i++)
	{
		if((i% columns)==0)
		{
			if(i>0 )
                            c+=sys_sprintf(&buf[c],"\n");
			c+=sys_sprintf(&buf[c],"%02x: ",i);
		}
		c+=sys_sprintf(&buf[c],"%02x ",hdmi_read(adv7619_cxt,i));
	}
	c+=sys_sprintf(&buf[c],"\n");
  return c;
}

static unsigned int adv7619hdmi_store(void *cxt, const char *buf, unsigned int count)
{
		adv7619_cxt_t *adv7619_cxt=cxt;
		unsigned reg;
		unsigned val;

		if(sys_sscanf(buf,"%x=%x", &reg,&val)==2)
		{
			debug_msg("set reg %x = %x\n",reg,val);
			hdmi_write(adv7619_cxt,reg,val);
		}
    return count;
}

static unsigned int adv7619afe_show(void *cxt,  char *buf)
{
	adv7619_cxt_t *adv7619_cxt=cxt;
	int i;
	int c;
	int columns=16;

	c=0;
	c+=sys_sprintf(&buf[c],"    ");
	for(i=0;i<columns;i++)
		c+=sys_sprintf(&buf[c],"%02x ",i);
	c+=sys_sprintf(&buf[c],"\n");

	c+=sys_sprintf(&buf[c],"____");
	for(i=0;i<columns;i++)
			c+=sys_sprintf(&buf[c],"___");
	c+=sys_sprintf(&buf[c],"\n");

	for(i=0;i<0x100;i++)
	{
		if((i% columns)==0)
		{
			if(i>0 )
                            c+=sys_sprintf(&buf[c],"\n");
			c+=sys_sprintf(&buf[c],"%02x: ",i);
		}
		c+=sys_sprintf(&buf[c],"%02x ",dpll_read(adv7619_cxt,i));
	}
	c+=sys_sprintf(&buf[c],"\n");
  return c;
}

static unsigned int adv7619afe_store(void *cxt, const char *buf, unsigned int count)
{
		adv7619_cxt_t *adv7619_cxt=cxt;
		unsigned reg;
		unsigned val;

		if(sys_sscanf(buf,"%x=%x", &reg,&val)==2)
		{
			debug_msg("set reg %x = %x\n",reg,val);
			dpll_write(adv7619_cxt,reg,val);
		}
    return count;
}

void adv7619_add_trace(handle_t adv7619_handle,void *trace_handle)
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    trace_attr_setup_t setup;
    trace_attr_handle_t attr_handle;
    char name[128];
    int index=0;
    enum
    {
        NO_ERROR,
        ERR_ADD_ATTR_IO,
        ERR_ADD_ATTR_CP,
        ERR_ADD_ATTR_HDMI,
        ERR_ADD_ATTR_AFE,
    }err=NO_ERROR;

    if(!adv7619_cxt || !trace_handle)
            return;


    do
    {
        sys_sprintf(name,"%s_%d_io",ADV7619_DRVNAME,adv7619_cxt->id);
        setup.name = name;
        setup.show = adv7619io_show;
        setup.store = adv7619io_store;
        setup.cxt = adv7619_cxt;
        attr_handle=trace_attr_create(&setup);
        if(!attr_handle)
        {
            err	=ERR_ADD_ATTR_IO;
            break;
        }
        trace_model_add_attr(trace_handle,attr_handle);
        adv7619_cxt->attr_handles[index++]=attr_handle;
        sys_sprintf(name,"%s_%d_cp",ADV7619_DRVNAME,adv7619_cxt->id);
        setup.name = name;
        setup.show = adv7619cp_show;
        setup.store = adv7619cp_store;
        setup.cxt = adv7619_cxt;
        attr_handle=trace_attr_create(&setup);
        if(!attr_handle)
        {
            err	=ERR_ADD_ATTR_CP;
            break;
        }
        trace_model_add_attr(trace_handle,attr_handle);
        adv7619_cxt->attr_handles[index++]=attr_handle;
        sys_sprintf(name,"%s_%d_hdmi",ADV7619_DRVNAME,adv7619_cxt->id);
        setup.name = name;
        setup.show = adv7619hdmi_show;
        setup.store = adv7619hdmi_store;
        setup.cxt = adv7619_cxt;
        attr_handle=trace_attr_create(&setup);
        if(!attr_handle)
        {
            err	=ERR_ADD_ATTR_HDMI;
            break;
        }
        trace_model_add_attr(trace_handle,attr_handle);
        adv7619_cxt->attr_handles[index++]=attr_handle;
        sys_sprintf(name,"%s_%d_afe",ADV7619_DRVNAME,adv7619_cxt->id);
        setup.name = name;
        setup.show = adv7619afe_show;
        setup.store = adv7619afe_store;
        setup.cxt = adv7619_cxt;
        attr_handle=trace_attr_create(&setup);
        if(!attr_handle)
        {
            err	=ERR_ADD_ATTR_AFE;
            break;
        }
        trace_model_add_attr(trace_handle,attr_handle);
        adv7619_cxt->attr_handles[index++]=attr_handle;
            
    }while(0);
}

static void adv7619_audiorate_detection(adv7619_cxt_t *adv7619_cxt,enum adv7619_audio_sample *audioinfo)
{
	//adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	//adv7619_cxt_t *state = adv7619_cxt;
	
	int cts = 0, n = 0;
	int tmds;
	int freq;
	int i;
	long diff, min;
	int table[] = {
		AUDIO_32K,
		AUDIO_44K,
		AUDIO_48K,
		AUDIO_88K,
		AUDIO_96K,
		AUDIO_176K,
		AUDIO_192K
	};
	int table_size;
	

	if(adv7619_cxt->lock != ADV7619_LOCK){
		*audioinfo = adv7619_cxt->rates = 4;    // 96kHz is used to mean unknown rate
		printk("%s return\n",__func__);
		return;
	}


	table_size = sizeof(table)/sizeof(table[0]);
	//hdmi_mask_write(adv7619_cxt, 0x5a, 0x01, 0x01);
	cts = hdmi_read(adv7619_cxt, 0x5b) << 12;
	cts |= hdmi_read(adv7619_cxt, 0x5c) << 4;
	cts |= ((hdmi_read(adv7619_cxt, 0x5d) >> 4) & 0x0f);

       
	n = (hdmi_read(adv7619_cxt, 0x5d) & 0x0f) << 16;
	n |= hdmi_read(adv7619_cxt, 0x5e) << 8;
	n |= hdmi_read(adv7619_cxt, 0x5f);

	//tmds = hdmi_read(adv7619_cxt, 0x06);
	
	tmds = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; 

	//printk("cts = %d, n = %d, tmds = %d\n",cts, n, tmds);


	if(cts == 0){
		*audioinfo = adv7619_cxt->rates = 4;    // 96kHz is used to mean unknown rate
		return;
	}


	freq = (((((long)tmds * n * 10) / cts) * 1000) >> 7)/1000;

	//printk("audio sampling freq = %d\n", freq);
    adv7619_cxt->rates = AUDIO_48K;
	for(i = 0, min = 0xff; i < table_size; i++)
	{
		diff = abs(freq - table[i]);
		if(diff <= min)
		{
			min = diff;
			adv7619_cxt->rates = i;
		}
	}
	*audioinfo = adv7619_cxt->rates;
	//printk("%s>>>>sample_rate = %d\n",__func__,adv7619_cxt->rates);
}

static void adv7619_enable_protection(adv7619_cxt_t *adv7619_cxt, BOOL_T enable)
{

	if(enable)
	{
		//printk("%s HDMI enable protection\n", __func__);
		adv7619_cxt->frameinfo.adv7619_hdcp_flag = 1;
		hdmi_write(adv7619_cxt, 0x1a, 0x1a);
		cp_write(adv7619_cxt, 0xbf, 0x05);
	}
	else
	{
		//printk("%s HDMI disable protection\n", __func__);
		adv7619_cxt->frameinfo.adv7619_hdcp_flag = 0;
		hdmi_write(adv7619_cxt, 0x1a, 0x0a);
		cp_write(adv7619_cxt, 0xbf, 0x04);
	}
}

void adv7619_get_audioinfoframe(handle_t adv7619_handle,U8_T *audioinfo) 
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    
    memcpy(audioinfo,adv7619_cxt->audio_infoframe,sizeof(adv7619_cxt->audio_infoframe)); 
}

static void adv7619_get_audioframe(adv7619_cxt_t *adv7619_cxt)
{
    U8_T InfoFrame[INFOFRAME_MAXSIZE]; 
    int i;
        
    mem_model_memset(adv7619_cxt->audio_infoframe,0,sizeof(adv7619_cxt->audio_infoframe));
       
		
    for(i = 0; i<3; i++)
        InfoFrame[i] = infoframe_read(adv7619_cxt, 0xe3+i); 
            
    for(i = 0; i<13; i++)
        InfoFrame[3+i] = infoframe_read(adv7619_cxt, 0x1d+i);  //0x1c check sum
    
    
    memcpy(adv7619_cxt->audio_infoframe,InfoFrame,sizeof(adv7619_cxt->audio_infoframe)); 
    
#if 0        
    {
        int jj;
	    printk(" =========================adv7619_get_audio============================\n");
	    //check audio infoframe data
	    for (jj=0; jj< INFOFRAME_MAXSIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            adv7619_cxt->audio_infoframe[jj-7],adv7619_cxt->audio_infoframe[jj-6],adv7619_cxt->audio_infoframe[jj-5],adv7619_cxt->audio_infoframe[jj-4],
	            adv7619_cxt->audio_infoframe[jj-3],adv7619_cxt->audio_infoframe[jj-2],adv7619_cxt->audio_infoframe[jj-1],adv7619_cxt->audio_infoframe[jj]);
		    }
	    }            
    }
#endif        
}

void adv7619_get_vendorinfoframe(handle_t adv7619_handle,U8_T *vendorinfo) 
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    
    memcpy(vendorinfo,adv7619_cxt->vendor_infoframe,sizeof(adv7619_cxt->vendor_infoframe)); 
}

static void adv7619_get_vendor(adv7619_cxt_t *adv7619_cxt)
{
    U8_T InfoFrame[INFOFRAME_MAXSIZE];
    int i;
        
    mem_model_memset(adv7619_cxt->vendor_infoframe,0,sizeof(adv7619_cxt->vendor_infoframe));
       
		
    for(i = 0; i<3; i++)
        InfoFrame[i] = infoframe_read(adv7619_cxt, 0xec+i); 
            
    for(i = 0; i<27; i++)
        InfoFrame[3+i] = infoframe_read(adv7619_cxt, 0x55+i);  //0x54 check sum
    
    
    memcpy(adv7619_cxt->vendor_infoframe,InfoFrame,sizeof(adv7619_cxt->vendor_infoframe)); 
    
#if 0        
    {
        int jj;
	    printk(" =========================adv7619_get_vendor============================\n");
	    //check vendor infoframe data
	    for (jj=0; jj< INFOFRAME_MAXSIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            adv7619_cxt->vendor_infoframe[jj-7],adv7619_cxt->vendor_infoframe[jj-6],adv7619_cxt->vendor_infoframe[jj-5],adv7619_cxt->vendor_infoframe[jj-4],
	            adv7619_cxt->vendor_infoframe[jj-3],adv7619_cxt->vendor_infoframe[jj-2],adv7619_cxt->vendor_infoframe[jj-1],adv7619_cxt->vendor_infoframe[jj]);
		    }
	    }            
    }
#endif        
}

void adv7619_get_aviinfoframe(handle_t adv7619_handle,U8_T *aviinfo) 
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    
    memcpy(aviinfo,adv7619_cxt->avi_infoframe,sizeof(adv7619_cxt->avi_infoframe)); 
}

static void adv7619_get_avi(adv7619_cxt_t *adv7619_cxt)
{
    U8_T InfoFrame[INFOFRAME_MAXSIZE]; 
    int i;
    int sampling_mode;
    int packet_colorspace;
        
    mem_model_memset(adv7619_cxt->avi_infoframe,0,sizeof(adv7619_cxt->avi_infoframe));
       
		
    for(i = 0; i<3; i++)
        InfoFrame[i] = infoframe_read(adv7619_cxt, 0xe0+i); 
            
    for(i = 0; i<16-3; i++)
        InfoFrame[3+i] = infoframe_read(adv7619_cxt, 0x01+i);  //0x00 check sum
            
    sampling_mode = (InfoFrame[3]>>5) & 0x03; //0:RGB 1:422 2:444 
        
    packet_colorspace = hdmi_read(adv7619_cxt, 0x53) & 0x0f;    
      
    memcpy(adv7619_cxt->avi_infoframe,InfoFrame,sizeof(adv7619_cxt->avi_infoframe)); 
    
#if 0        
    {
        int jj;
	    printk(" =========================adv7619_get_avi============================\n");
	    //check avi infoframe data
	    for (jj=0; jj< INFOFRAME_MAXSIZE; jj++)
	    {
		    if (jj%8 ==7)
	 	    {
			  
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            adv7619_cxt->avi_infoframe[jj-7],adv7619_cxt->avi_infoframe[jj-6],adv7619_cxt->avi_infoframe[jj-5],adv7619_cxt->avi_infoframe[jj-4],
	            adv7619_cxt->avi_infoframe[jj-3],adv7619_cxt->avi_infoframe[jj-2],adv7619_cxt->avi_infoframe[jj-1],adv7619_cxt->avi_infoframe[jj]);
		    }
	    } 
	    if (sampling_mode ==0)
	        printk("avi info RGB\n"); 
	    if (sampling_mode ==1)
	        printk("avi info YUV422\n");  
	    if (sampling_mode ==2)
	        printk("avi info YUV444\n");             
    }
#endif        
}

static void adv7619_get_videoinfo(adv7619_cxt_t *adv7619_cxt,video_timing_t *timing)
{
    
    if(timing)
    {
        U32_T freq;
        U32_T htotal,vtotal;
        U32_T dInput;
        U8_T InfoFrame[30]; 
        int i;
        int sampling_mode;
        int packet_colorspace;
        int color_mode;
        U64_T freq_1000;
        U32_T denominator;
        
        //sys_msleep(50);
        mem_model_memset(timing,0,sizeof(*timing));
        timing->width = ((hdmi_read(adv7619_cxt, 0x07) & 0x1f) * 0x100)+ hdmi_read(adv7619_cxt, 0x08);
	    timing->height = ((hdmi_read(adv7619_cxt, 0x09) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x0a);
	    timing->interlaced = ((hdmi_read(adv7619_cxt, 0x0b) & 0x20)==0x20);
	    freq = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; //TMDS frequency measurement
	    
	    if (hdmi_read(adv7619_cxt, 0x5) & 0x80) //1 - HDMI Mode Detected 
 	    {
		/* adjust for deep color mode */
		    unsigned bits_per_channel = ((hdmi_read(adv7619_cxt, 0x0b) & 0xc0) >> 5) + 8; //8 or 10 or 12 bits
		    freq = freq * 8 / bits_per_channel;
		    //printk("%s deep color %dx pixclock %d\n",__func__,bits_per_channel,freq);
	    }
	    timing->detail.pixelclock = freq;
        
	    timing->detail.hfrontporch = ((hdmi_read(adv7619_cxt, 0x20) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x21);
	    timing->detail.hsync = ((hdmi_read(adv7619_cxt, 0x22) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x23);
	    timing->detail.hbackporch = ((hdmi_read(adv7619_cxt, 0x24) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x25);
	    timing->detail.vfrontporch = (((hdmi_read(adv7619_cxt, 0x2a) & 0x3f) * 0x100) + hdmi_read(adv7619_cxt, 0x2b)) / 2;
	    timing->detail.vsync = (((hdmi_read(adv7619_cxt, 0x2e) & 0x3f) * 0x100) + hdmi_read(adv7619_cxt, 0x2f)) / 2;
	    timing->detail.vbackporch = (((hdmi_read(adv7619_cxt, 0x32) & 0x3f) * 0x100) + 	hdmi_read(adv7619_cxt, 0x33)) / 2;
	    timing->detail.polarities = ((hdmi_read(adv7619_cxt, 0x05) & 0x10) ? VSYNC_POS_MASK : 0) | ((hdmi_read(adv7619_cxt, 0x05) & 0x20) ? HSYNC_POS_MASK : 0);
	    if (timing->interlaced ) 
        {
            timing->height += ((hdmi_read(adv7619_cxt, 0x0b) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x0c);
            timing->detail.il_vfrontporch = (((hdmi_read(adv7619_cxt, 0x2c) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x2d)) / 2;
            timing->detail.il_vsync = (((hdmi_read(adv7619_cxt, 0x30) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x31)) / 2;
            timing->detail.il_vbackporch = (((hdmi_read(adv7619_cxt, 0x34) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x35)) / 2;
	    }
        htotal=timing->width
               +timing->detail.hfrontporch+timing->detail.hsync+timing->detail.hbackporch;
        vtotal=timing->height+
               +timing->detail.vfrontporch+timing->detail.vsync+timing->detail.vbackporch
               +timing->detail.il_vfrontporch + timing->detail.il_vsync + timing->detail.il_vbackporch; 
        if(htotal*vtotal)
        {
		    dInput = ((freq/htotal)*1000)/vtotal + 500;
		    //printk("%s timing->framerate *1000=%d\n",__func__,(dInput-500));
            timing->framerate = dInput/1000;
            //printk("%s htotal %d vtotal %d; timing->framerate=%d\n",__func__,htotal,vtotal,timing->framerate);
		}
        //printk("%s %dx%d%c pixclock %d\n",__func__,timing->width,(timing->interlaced) ? (timing->height)/2:timing->height,(timing->interlaced) ? 'i':'p',freq);
        
        if((timing->detail.pixelclock >170000000) || ((timing->width== 2560) && (timing->height == 1080))) //special case
        {
			adv7619_cxt->frameinfo.dual_pixel=1;
            adv7619_cxt->frameinfo.dual_pixel_like=1;	
		}
		else
		{
			adv7619_cxt->frameinfo.dual_pixel=0;
            adv7619_cxt->frameinfo.dual_pixel_like=0;	
		}
		
		for(i = 0; i<3; i++)
            InfoFrame[i] = infoframe_read(adv7619_cxt, 0xe0+i); 
            
        for(i = 0; i<16-3; i++)
            InfoFrame[3+i] = infoframe_read(adv7619_cxt, 0x01+i);  
            
        sampling_mode = (InfoFrame[3]>>5) & 0x03; //0:RGB 1:422 2:444 
        
        color_mode = (InfoFrame[4]>>6) & 0x03; //0:NONE 1:ITU601 2:ITU709 
        
        packet_colorspace = hdmi_read(adv7619_cxt, 0x53) & 0x0f;
#if 1        
        if (packet_colorspace ==0)
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_RGB_LIMIT;
		}
        else if (packet_colorspace ==1 || packet_colorspace ==0x0a)
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_RGB_FULL;  
		} 
        else
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_YUV;  
		}
#else
		adv7619_cxt->frameinfo.packet_colorspace = CS_YUV; //fix to YUV packet colorspace
#endif 
        if (color_mode == 1)
        {
            adv7619_cxt->frameinfo.is_ITU601 = 1;
            adv7619_cxt->frameinfo.is_ITU709 = 0;
            //printk("====HDMI ITU601\n");
		}
        else if (color_mode == 2)
        { 
            adv7619_cxt->frameinfo.is_ITU601 = 0;
            adv7619_cxt->frameinfo.is_ITU709 = 1;
            //printk("====HDMI ITU709\n");
		}
        else
        {
            adv7619_cxt->frameinfo.is_ITU601 = 0;
            adv7619_cxt->frameinfo.is_ITU709 = 0;  
            //printk("====HDMI is not standard ITU\n");  
		}
		
		if (sampling_mode == 0)
        {
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 1;
            //printk("====HDMI RGB\n");
		}
		else if (sampling_mode == 1)
        {
            adv7619_cxt->frameinfo.is_yuv422 = 1;
            adv7619_cxt->frameinfo.is_yuv444 = 0;
            //printk("====HDMI YUV422\n");
		}
        else if (sampling_mode == 2)
        { 
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 1;
            //printk("====HDMI YUV444\n");
		}
        else
        {
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 0;
            //printk("====HDMI sampling_mode not ready \n");  
		}     
        adv7619_cxt->frameinfo.sampling_mode = sampling_mode;
        adv7619_cxt->frameinfo.width = timing->width;
        adv7619_cxt->frameinfo.height = timing->height;
        adv7619_cxt->frameinfo.is_interlace = timing->interlaced;
        adv7619_cxt->frameinfo.pixel_clock = freq;
     
        adv7619_cxt->frameinfo.framerate = timing->framerate;
        freq = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; 
        freq_1000 = freq*1000;
        
        adv7619_cxt->frameinfo.denominator = ((freq/htotal)*1000)/vtotal; 
        
        if ((adv7619_cxt->frameinfo.denominator < 30100) && (adv7619_cxt->frameinfo.denominator > 29992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 30000;
		}
        
        if ((adv7619_cxt->frameinfo.denominator < 29993) && (adv7619_cxt->frameinfo.denominator > 29900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 29970;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 60100) && (adv7619_cxt->frameinfo.denominator > 59992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 60000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 59993) && (adv7619_cxt->frameinfo.denominator > 59900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 59940;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 50100) && (adv7619_cxt->frameinfo.denominator > 49992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 50000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 24100) && (adv7619_cxt->frameinfo.denominator > 23995))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 24000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 23995) && (adv7619_cxt->frameinfo.denominator > 23900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 23980;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 25100) && (adv7619_cxt->frameinfo.denominator > 24900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 25000;
		}
        
        //printk("%s frameinfo.denominator %d (%d)\n",__func__,adv7619_cxt->frameinfo.denominator,denominator);  
         
        //printk("======HDMI or DVI mode detection: %s\n", hdmi_read(adv7619_cxt, 0x05) & 0x80 ? "HDMI" : "DVI");
      
        //printk("======HDMI color space(0,1=RGB,others=YUV): %d\n", hdmi_read(adv7619_cxt, 0x53) & 0x0f);
    }
}

static void adv7619_get_timing(adv7619_cxt_t *adv7619_cxt,video_timing_t *timing)
{
    
    if(timing)
    {
        U32_T freq;
        U32_T htotal,vtotal;
        U32_T dInput;
        U8_T InfoFrame[30]; 
        int i;
        int sampling_mode;
        int packet_colorspace;
        int color_mode;
        U64_T freq_1000;
        U32_T denominator;
        
        //sys_msleep(50);
        mem_model_memset(timing,0,sizeof(*timing));
        timing->width = ((hdmi_read(adv7619_cxt, 0x07) & 0x1f) * 0x100)+ hdmi_read(adv7619_cxt, 0x08);
	    timing->height = ((hdmi_read(adv7619_cxt, 0x09) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x0a);
	    timing->interlaced = ((hdmi_read(adv7619_cxt, 0x0b) & 0x20)==0x20);
	    freq = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; //TMDS frequency measurement
	    
	    if (hdmi_read(adv7619_cxt, 0x5) & 0x80) //1 - HDMI Mode Detected 
 	    {
		/* adjust for deep color mode */
		    unsigned bits_per_channel = ((hdmi_read(adv7619_cxt, 0x0b) & 0xc0) >> 5) + 8; //8 or 10 or 12 bits
		    freq = freq * 8 / bits_per_channel;
		    //printk("%s deep color %dx pixclock %d\n",__func__,bits_per_channel,freq);
	    }
	    timing->detail.pixelclock = freq;
        
	    timing->detail.hfrontporch = ((hdmi_read(adv7619_cxt, 0x20) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x21);
	    timing->detail.hsync = ((hdmi_read(adv7619_cxt, 0x22) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x23);
	    timing->detail.hbackporch = ((hdmi_read(adv7619_cxt, 0x24) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x25);
	    timing->detail.vfrontporch = (((hdmi_read(adv7619_cxt, 0x2a) & 0x3f) * 0x100) + hdmi_read(adv7619_cxt, 0x2b)) / 2;
	    timing->detail.vsync = (((hdmi_read(adv7619_cxt, 0x2e) & 0x3f) * 0x100) + hdmi_read(adv7619_cxt, 0x2f)) / 2;
	    timing->detail.vbackporch = (((hdmi_read(adv7619_cxt, 0x32) & 0x3f) * 0x100) + 	hdmi_read(adv7619_cxt, 0x33)) / 2;
	    timing->detail.polarities = ((hdmi_read(adv7619_cxt, 0x05) & 0x10) ? VSYNC_POS_MASK : 0) | ((hdmi_read(adv7619_cxt, 0x05) & 0x20) ? HSYNC_POS_MASK : 0);
	    if (timing->interlaced ) 
        {
            timing->height += ((hdmi_read(adv7619_cxt, 0x0b) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x0c);
            timing->detail.il_vfrontporch = (((hdmi_read(adv7619_cxt, 0x2c) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x2d)) / 2;
            timing->detail.il_vsync = (((hdmi_read(adv7619_cxt, 0x30) & 0x1f) * 0x100)  + hdmi_read(adv7619_cxt, 0x31)) / 2;
            timing->detail.il_vbackporch = (((hdmi_read(adv7619_cxt, 0x34) & 0x1f) * 0x100) + hdmi_read(adv7619_cxt, 0x35)) / 2;
	    }
        htotal=timing->width
               +timing->detail.hfrontporch+timing->detail.hsync+timing->detail.hbackporch;
        vtotal=timing->height+
               +timing->detail.vfrontporch+timing->detail.vsync+timing->detail.vbackporch
               +timing->detail.il_vfrontporch + timing->detail.il_vsync + timing->detail.il_vbackporch; 
        if(htotal*vtotal)
        {
		    dInput = ((freq/htotal)*1000)/vtotal + 500;
		    //printk("%s timing->framerate *1000=%d\n",__func__,(dInput-500));
            timing->framerate = dInput/1000;
            //printk("%s htotal %d vtotal %d; timing->framerate=%d\n",__func__,htotal,vtotal,timing->framerate);
		}
        //printk("%s %dx%d%c pixclock %d\n",__func__,timing->width,(timing->interlaced) ? (timing->height)/2:timing->height,(timing->interlaced) ? 'i':'p',freq);
        
        if((timing->detail.pixelclock >170000000) || ((timing->width== 2560) && (timing->height == 1080))) //special case
        {
			adv7619_cxt->frameinfo.dual_pixel=1;
            adv7619_cxt->frameinfo.dual_pixel_like=1;	
		}
		else
		{
			adv7619_cxt->frameinfo.dual_pixel=0;
            adv7619_cxt->frameinfo.dual_pixel_like=0;	
		}
		
		for(i = 0; i<3; i++)
            InfoFrame[i] = infoframe_read(adv7619_cxt, 0xe0+i); 
            
        for(i = 0; i<16-3; i++)
            InfoFrame[3+i] = infoframe_read(adv7619_cxt, 0x01+i);  
            
        sampling_mode = (InfoFrame[3]>>5) & 0x03; //0:RGB 1:422 2:444 
        
        color_mode = (InfoFrame[4]>>6) & 0x03; //0:NONE 1:ITU601 2:ITU709 
        
        packet_colorspace = hdmi_read(adv7619_cxt, 0x53) & 0x0f;
#if 1        
        if (packet_colorspace ==0)
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_RGB_LIMIT;
		}
        else if (packet_colorspace ==1 || packet_colorspace ==0x0a)
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_RGB_FULL;  
		} 
        else
        {
            adv7619_cxt->frameinfo.packet_colorspace = CS_YUV;  
		}
#else
		adv7619_cxt->frameinfo.packet_colorspace = CS_YUV; //fix to YUV packet colorspace
#endif 
        if (color_mode == 1)
        {
            adv7619_cxt->frameinfo.is_ITU601 = 1;
            adv7619_cxt->frameinfo.is_ITU709 = 0;
            printk("====HDMI ITU601\n");
		}
        else if (color_mode == 2)
        { 
            adv7619_cxt->frameinfo.is_ITU601 = 0;
            adv7619_cxt->frameinfo.is_ITU709 = 1;
            printk("====HDMI ITU709\n");
		}
        else
        {
            adv7619_cxt->frameinfo.is_ITU601 = 0;
            adv7619_cxt->frameinfo.is_ITU709 = 0;  
            printk("====HDMI is not standard ITU\n");  
		}
		
		if (sampling_mode == 0)
        {
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 1;
            printk("====HDMI RGB\n");
		}
		else if (sampling_mode == 1)
        {
            adv7619_cxt->frameinfo.is_yuv422 = 1;
            adv7619_cxt->frameinfo.is_yuv444 = 0;
            printk("====HDMI YUV422\n");
		}
        else if (sampling_mode == 2)
        { 
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 1;
            printk("====HDMI YUV444\n");
		}
        else
        {
            adv7619_cxt->frameinfo.is_yuv422 = 0;
            adv7619_cxt->frameinfo.is_yuv444 = 0;
            printk("====HDMI sampling_mode not ready \n");  
		}     
        adv7619_cxt->frameinfo.sampling_mode = sampling_mode;
        adv7619_cxt->frameinfo.width = timing->width;
        adv7619_cxt->frameinfo.height = timing->height;
        adv7619_cxt->frameinfo.is_interlace = timing->interlaced;
        adv7619_cxt->frameinfo.pixel_clock = freq;
     
        adv7619_cxt->frameinfo.framerate = timing->framerate;
        freq = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; 
        freq_1000 = freq*1000;
        
        adv7619_cxt->frameinfo.denominator = ((freq/htotal)*1000)/vtotal; 
        
        if ((adv7619_cxt->frameinfo.denominator < 30100) && (adv7619_cxt->frameinfo.denominator > 29992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 30000;
		}
        
        if ((adv7619_cxt->frameinfo.denominator < 29993) && (adv7619_cxt->frameinfo.denominator > 29900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 29970;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 60100) && (adv7619_cxt->frameinfo.denominator > 59992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 60000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 59993) && (adv7619_cxt->frameinfo.denominator > 59900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 59940;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 50100) && (adv7619_cxt->frameinfo.denominator > 49992))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 50000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 24100) && (adv7619_cxt->frameinfo.denominator > 23995))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 24000;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 23995) && (adv7619_cxt->frameinfo.denominator > 23900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 23980;
		}
		
		if ((adv7619_cxt->frameinfo.denominator < 25100) && (adv7619_cxt->frameinfo.denominator > 24900))
        {
			denominator = adv7619_cxt->frameinfo.denominator;
			adv7619_cxt->frameinfo.denominator = 25000;
		}
        
        printk("%s frameinfo.denominator %d (%d)\n",__func__,adv7619_cxt->frameinfo.denominator,denominator);  
         
        printk("======HDMI or DVI mode detection: %s\n", hdmi_read(adv7619_cxt, 0x05) & 0x80 ? "HDMI" : "DVI");
      
        printk("======HDMI color space(0,1=RGB,others=YUV): %d\n", hdmi_read(adv7619_cxt, 0x53) & 0x0f);
        
        if (adv7619_cxt->frameinfo.width == 3840)
        {
			if (adv7619_cxt->frameinfo.height == 2160)
			    adv7619_cxt->desired_timing = 1;
			else
			    adv7619_cxt->desired_timing = 0;
			    
			if ((denominator < 30160) && (denominator > 29810))  
			    adv7619_cxt->desired_framerate = 1;
			else if ((denominator < 25130) && (denominator > 24860))  
			    adv7619_cxt->desired_framerate = 1;
			else if ((denominator < 24130) && (denominator > 23840))  
			    adv7619_cxt->desired_framerate = 1;
			else        
			    adv7619_cxt->desired_framerate = 0;
		}
		else if (adv7619_cxt->frameinfo.width == 1920)
		{
			if ((adv7619_cxt->frameinfo.height == 1080) /*|| (adv7619_cxt->frameinfo.height == 540)*/)
			    adv7619_cxt->desired_timing = 1;
			else
			    adv7619_cxt->desired_timing = 0;
			    
			if ((denominator < 60310) && (denominator > 59630))  
			    adv7619_cxt->desired_framerate = 1;
			else if ((denominator < 50260) && (denominator > 49740))  
			    adv7619_cxt->desired_framerate = 1;
			else        
			    adv7619_cxt->desired_framerate = 0;    
		}
		else if (adv7619_cxt->frameinfo.width == 1280)
		{
			if (adv7619_cxt->frameinfo.height == 720)
			    adv7619_cxt->desired_timing = 1;
			else
			    adv7619_cxt->desired_timing = 0;
			    
			if ((denominator < 60310) && (denominator > 59630))  
			    adv7619_cxt->desired_framerate = 1;
			else if ((denominator < 50260) && (denominator > 49740))  
			    adv7619_cxt->desired_framerate = 1;
			else        
			    adv7619_cxt->desired_framerate = 0;        
		}
		else if (adv7619_cxt->frameinfo.width == 720)
		{
			if ((adv7619_cxt->frameinfo.height == 480) || (adv7619_cxt->frameinfo.height == 576))
			    adv7619_cxt->desired_timing = 1;
			else
			    adv7619_cxt->desired_timing = 0;
			    
			if ((denominator < 60310) && (denominator > 59630))  
			    adv7619_cxt->desired_framerate = 1;
			else if ((denominator < 50260) && (denominator > 49740))  
			    adv7619_cxt->desired_framerate = 1;
			else        
			    adv7619_cxt->desired_framerate = 0;            
		}
		//else if (adv7619_cxt->frameinfo.width == 720)
		//{
		//	if ((adv7619_cxt->frameinfo.height == 576) || (adv7619_cxt->frameinfo.height == 288))
		//	    adv7619_cxt->desired_timing = 1;
		//	else
		//	    adv7619_cxt->desired_timing = 0;
		//}
		else if (adv7619_cxt->frameinfo.width == 640)
		{
			if (adv7619_cxt->frameinfo.height == 480)
			    adv7619_cxt->desired_timing = 1;
			else
			    adv7619_cxt->desired_timing = 0;
			    
			if ((denominator < 60310) && (denominator > 59630))  
			    adv7619_cxt->desired_framerate = 1;
			else        
			    adv7619_cxt->desired_framerate = 0;     
		}
		else
	    {
		    adv7619_cxt->desired_timing = 0;
		    adv7619_cxt->desired_framerate = 0;  
		}
    }
}

static void adv7619_task(void *data)
{
    adv7619_cxt_t *adv7619_cxt=data;
    int i;
    int stat;
    int raw;
    U32_T last_width,last_height,last_framerate;
	BOOL_T is_interlace;
    adv7619_out_color_ctrl_t out_color_setting;
    video_timing_t timing;
    U32_T pixel_clock;
    enum adv7619_audio_sample audio_rate_detection;
    int last_audio_rate;           
    regtbl_item_t *setting=NULL;
    int setting_count=0;
    int status_05;
    
    stat=io_read(adv7619_cxt,0x5c); 
	raw=io_read(adv7619_cxt,0x5b); 
	
#if 1
    status_05=hdmi_read(adv7619_cxt,0x05);
  
    if(status_05 & 0x40)
    {          
        //if (adv7619_cxt->lock != ADV7619_HDCP)
        {
            adv7619_enable_protection(adv7619_cxt, 1);
            adv7619_callback(adv7619_cxt,ADV7619_HDCP);
            adv7619_cxt->lock = ADV7619_HDCP;
		}
    }
    else
    {
		//if (adv7619_cxt->lock == ADV7619_HDCP)
		{
		    adv7619_enable_protection(adv7619_cxt, 0);
		    adv7619_callback(adv7619_cxt,ADV7619_HDCP);
		}
	}
    
#endif

#if 1
    if (adv7619_cxt->lock == ADV7619_LOCK)
    {
		last_audio_rate = adv7619_cxt->detected_audiorate; 
        adv7619_audiorate_detection(adv7619_cxt,&audio_rate_detection);
        adv7619_cxt->detected_audiorate = audio_rate_detection;
        if (last_audio_rate != audio_rate_detection)
        {
			adv7619_callback(adv7619_cxt,ADV7619_AUDIO_RATE);
		} 
	}
#endif
 
#if 1
	last_width = adv7619_cxt->frameinfo.width;
    last_height = adv7619_cxt->frameinfo.height; 
    last_framerate = adv7619_cxt->frameinfo.framerate;
	is_interlace = adv7619_cxt->frameinfo.is_interlace;
	
	if ((last_width ==0) && (last_height ==0))
	{
	    //printk(">>reset signal changed\n");
	    adv7619_get_videoinfo(adv7619_cxt,&timing);
	    last_width = adv7619_cxt->frameinfo.width;
        last_height = adv7619_cxt->frameinfo.height; 
        last_framerate = adv7619_cxt->frameinfo.framerate;
	    is_interlace = adv7619_cxt->frameinfo.is_interlace;
	    return;
    }
             
    adv7619_get_videoinfo(adv7619_cxt,&timing);

#if 1
    if (adv7619_cxt->lock != ADV7619_UNLOCK_ENHANCE) {
        adv7619_cxt->before_frame.is_available = FALSE;
    }
#endif

	if (adv7619_cxt->frameinfo.width != last_width || 
		adv7619_cxt->frameinfo.height != last_height || 
		adv7619_cxt->frameinfo.framerate != last_framerate ||
		adv7619_cxt->frameinfo.is_interlace != is_interlace)
	{
#if 1
		printk("[adv7619_task:%p] w:%d -> %d, h:%d -> %d, f:%d -> %d, is_i:%d -> %d\n",
		adv7619_cxt,
		last_width,     adv7619_cxt->frameinfo.width,
		last_height,    adv7619_cxt->frameinfo.height,
		last_framerate, adv7619_cxt->frameinfo.framerate,
		is_interlace,   adv7619_cxt->frameinfo.is_interlace );
#endif
		if ((adv7619_cxt->frameinfo.width ==0) && (adv7619_cxt->frameinfo.height ==0))
		    printk("signal changed not ready\n");  
					//cp_mask_write(adv7619_cxt,0xbf,0x05,0x05);
#if 1 // for the case of specific media player 
		
#if 1
        // 1frameinfo skip
        if(adv7619_cxt->lock == ADV7619_LOCK && raw & 0x02) {
            if(adv7619_cxt->skip_unlock_enhance == FALSE){
                printk("[adv7619_task:%p]skip frameinfo\n", adv7619_cxt);
                adv7619_cxt->skip_unlock_enhance = TRUE;
                adv7619_cxt->frameinfo.width = last_width;
                adv7619_cxt->frameinfo.height = last_height;
                adv7619_cxt->frameinfo.framerate = last_framerate;
                adv7619_cxt->frameinfo.is_interlace = is_interlace;
                return;
            } else {
                printk("[adv7619_task:%p]non skip frameinfo\n", adv7619_cxt);
            }
        }
        adv7619_cxt->skip_unlock_enhance = FALSE;
#endif

		if (adv7619_cxt->lock != ADV7619_UNLOCK_ENHANCE)
		{ 
			adv7619_callback(adv7619_cxt,ADV7619_UNLOCK_ENHANCE);
			adv7619_cxt->lock = ADV7619_UNLOCK_ENHANCE;
#if 1
            if(raw & 0x02) {
                adv7619_cxt->before_frame.is_available = TRUE;
                adv7619_cxt->before_frame.width = last_width;
                adv7619_cxt->before_frame.height = last_height;
                adv7619_cxt->before_frame.framerate = last_framerate;
                adv7619_cxt->before_frame.interlaced = is_interlace;
            }
#endif 
		}
#if 1
        else {
            if(raw & 0x02 &&
				adv7619_cxt->before_frame.is_available == TRUE &&
				adv7619_cxt->frameinfo.width == adv7619_cxt->before_frame.width &&
				adv7619_cxt->frameinfo.height == adv7619_cxt->before_frame.height &&
				adv7619_cxt->frameinfo.framerate == adv7619_cxt->before_frame.framerate &&
				adv7619_cxt->frameinfo.is_interlace == adv7619_cxt->before_frame.interlaced){
               
				adv7619_cxt->lock = ADV7619_LOCK;
                adv7619_cxt->skip_unlock_enhance = FALSE;
                adv7619_callback(adv7619_cxt,ADV7619_LOCK);
                hdmi_write(adv7619_cxt, 0x1a, 0x0a);
                cp_mask_write(adv7619_cxt,0xbf,0x05,0x04);
                io_mask_write(adv7619_cxt,0x5d,0x0a,stat);
                printk(">>>>>>>>>>>>>>>>>>>>>>adv7619 UNLOCK_ENHANCE -> LOCK\n");
            }
        }
#endif
#endif
		if ((adv7619_cxt->frameinfo.width ==3840) /*|| (last_width ==3840)*/)
		{
		    setting=Setting_4k2k;
            setting_count=sizeof(Setting_4k2k)/sizeof(regtbl_item_t);
            for(i=0;i<setting_count;i++)
            {
                adv7619_wr_map_reg(adv7619_cxt,setting[i].domain,setting[i].reg,setting[i].value);
            }
		}
		printk(".....adv7619 signal changed\n");
		//printk("frameinfo.width=%d, last_width=%d\n",adv7619_cxt->frameinfo.width,last_width);
		return;
	}
		//else
		//    adv7619_cxt->lock = ADV7619_OTHER;
				    
#endif
#if 1  
    adv7619_cxt->skip_unlock_enhance = FALSE;
#endif

    if(stat & 0x0f)
    {
		if(stat & 0x02)
		{
            raw=io_read(adv7619_cxt,0x5b); 
            if(raw & 0x02)
            {
				if (adv7619_cxt->lock == ADV7619_LOCK)
				{
					io_write(adv7619_cxt,0x5d,stat);
					return;
				}
				else
					printk("adv7619 signal changed noted\n");
					
                last_width = adv7619_cxt->frameinfo.width;
                last_height = adv7619_cxt->frameinfo.height; 
                last_framerate = adv7619_cxt->frameinfo.framerate;
				is_interlace = adv7619_cxt->frameinfo.is_interlace;
             
                adv7619_get_timing(adv7619_cxt,&timing);
                adv7619_get_avi(adv7619_cxt);
                adv7619_get_vendor(adv7619_cxt);
                adv7619_get_audioframe(adv7619_cxt);
				if (adv7619_cxt->frameinfo.width != last_width || 
					adv7619_cxt->frameinfo.height != last_height || 
					adv7619_cxt->frameinfo.framerate != last_framerate ||
					adv7619_cxt->frameinfo.is_interlace != is_interlace)
				{
					//cp_mask_write(adv7619_cxt,0xbf,0x05,0x05);
					#if 0 // for the case of specific media player 
					adv7619_callback(adv7619_cxt,ADV7619_UNLOCK);
					#endif
					printk("adv7619 signal changed and recheck...\n");
					return;
				}
				
				if ((adv7619_cxt->desired_timing != 1) || (adv7619_cxt->desired_framerate != 1))
				{
					if (adv7619_cxt->desired_timing != 1) 
		    		    printk(">>>>>>adv7619 timing not desired\n");
		    		if (adv7619_cxt->desired_framerate != 1) 
		    		    printk(">>>>>>adv7619 framerate not desired\n");  
		    		    
		    		adv7619_cxt->lock = ADV7619_UNLOCK;  
	        		adv7619_callback(adv7619_cxt,ADV7619_UNLOCK);
					hdmi_write(adv7619_cxt, 0x1a, 0x1a);
					cp_mask_write(adv7619_cxt,0xbf,0x05,0x05);
					io_mask_write(adv7619_cxt,0x5d,0x0a,stat);
					return;
				}                    
                if (timing.interlaced)
                    printk(">>>>adv7619 %d lock %dx%d%c%d\n",adv7619_cxt->id,timing.width,timing.height,(timing.interlaced) ? 'i':'p',timing.framerate*2);
                else
                    printk(">>>>adv7619 %d lock %dx%d%c%d\n",adv7619_cxt->id,timing.width,timing.height,(timing.interlaced) ? 'i':'p',timing.framerate);
                        
                pixel_clock=timing.detail.pixelclock;
                        
                out_color_setting.alt_gamma=0;
                out_color_setting.op_656_range=1;
                out_color_setting.rgb_out=0;
                out_color_setting.alt_data_sat=0;
                        
               if((pixel_clock>170000000) || (timing.width == 3840))
                {
		            printk("Setting_4k2k\n");
                    setting=Setting_4k2k;
                    setting_count=sizeof(Setting_4k2k)/sizeof(regtbl_item_t);
                }
                else if(pixel_clock>=74250000) 
                {
                    setting=Setting_HD;
                    setting_count=sizeof(Setting_HD)/sizeof(regtbl_item_t);
                }
                else
                {
                    setting=Setting_SD;
                    setting_count=sizeof(Setting_SD)/sizeof(regtbl_item_t);
                }
                       
                for(i=0;i<setting_count;i++)
                {
                    adv7619_wr_map_reg(adv7619_cxt,setting[i].domain,setting[i].reg,setting[i].value);
                }
                
                if (adv7619_cxt->frameinfo.sampling_mode == 0) // RGB transfer to BT709 //not care
                    out_color_setting.alt_gamma=1; 
                        
                adv7619_set_out_color(adv7619_cxt,&out_color_setting);
                
                adv7619_cxt->lock = ADV7619_LOCK;              
#if 1 
                adv7619_cxt->skip_unlock_enhance = FALSE;
#endif
                adv7619_callback(adv7619_cxt,ADV7619_LOCK);
				
				hdmi_write(adv7619_cxt, 0x1a, 0x0a);
                cp_mask_write(adv7619_cxt,0xbf,0x05,0x04);
                io_mask_write(adv7619_cxt,0x5d,0x0a,stat);
                printk(">>>>>>>>>>>>>>>>>>>>>>adv7619 lock\n");
			}
			else if (raw & 0x04) 
			{
				if (adv7619_cxt->lock != ADV7619_UNLOCK) 
				{
		    		printk(">>adv7619 unlock\n");
		    	    if (last_width ==3840)
		    		{
		    		    setting=Setting_4k2k;
                        setting_count=sizeof(Setting_4k2k)/sizeof(regtbl_item_t);
                        for(i=0;i<setting_count;i++)
                        {
                            adv7619_wr_map_reg(adv7619_cxt,setting[i].domain,setting[i].reg,setting[i].value);
                        }
					}
		    		
	        		adv7619_callback(adv7619_cxt,ADV7619_UNLOCK);
					adv7619_cxt->lock = ADV7619_UNLOCK;
					hdmi_write(adv7619_cxt, 0x1a, 0x1a);
					cp_mask_write(adv7619_cxt,0xbf,0x05,0x05);
					io_mask_write(adv7619_cxt,0x5d,0x0a,stat);
				}
			} 
        }
           
        //io_write(adv7619_cxt,0x5d,stat);
    }
}

void adv7619_get_audiorate(handle_t adv7619_handle,enum adv7619_audio_sample *audioinfo)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	//adv7619_cxt_t *state;
	
	int cts = 0, n = 0;
	int tmds;
	int freq;
	int i;
	long diff, min;
	int table[] = {
		AUDIO_32K,
		AUDIO_44K,
		AUDIO_48K,
		AUDIO_88K,
		AUDIO_96K,
		AUDIO_176K,
		AUDIO_192K
	};
	int table_size;
	

	if(adv7619_cxt->lock != ADV7619_LOCK){
		*audioinfo = adv7619_cxt->rates = 4;    // 96kHz
		printk("%s return\n",__func__);
		return;
	}


	table_size = sizeof(table)/sizeof(table[0]);
	hdmi_mask_write(adv7619_cxt, 0x5a, 0x01, 0x01);
	cts = hdmi_read(adv7619_cxt, 0x5b) << 12;
	cts |= hdmi_read(adv7619_cxt, 0x5c) << 4;
	cts |= ((hdmi_read(adv7619_cxt, 0x5d) >> 4) & 0x0f);

       
	n = (hdmi_read(adv7619_cxt, 0x5d) & 0x0f) << 16;
	n |= hdmi_read(adv7619_cxt, 0x5e) << 8;
	n |= hdmi_read(adv7619_cxt, 0x5f);

	//tmds = hdmi_read(adv7619_cxt, 0x06);
	
	tmds = ((( (hdmi_read(adv7619_cxt, 0x51) * 0x100)  + hdmi_read(adv7619_cxt, 0x52)  ))* (1000000 >> 6 )) >>1 ; 

	printk("cts = %d, n = %d, tmds = %d\n",cts, n, tmds);


	if(cts == 0){
		*audioinfo = adv7619_cxt->rates = 4;    // 96kHz
		return;
	}


	freq = (((((long)tmds * n * 10) / cts) * 1000) >> 7)/1000;

	printk("audio sampling freq = %d\n", freq/10);
    adv7619_cxt->rates = AUDIO_48K;
	for(i = 0, min = 0xff; i < table_size; i++)
	{
		diff = abs(freq - table[i]);
		if(diff <= min)
		{
			min = diff;
			adv7619_cxt->rates = i;
		}
	}
	*audioinfo = adv7619_cxt->rates;
	//printk(">>>>sample_rate = %d\n", adv7619_cxt->rates);
}

void adv7619_get_yuvmode(handle_t adv7619_handle,int *yuv422)
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
	
	if ((adv7619_cxt->frameinfo.is_yuv422 == 1) && (adv7619_cxt->frameinfo.is_yuv444 ==0))
	    *yuv422=1;
	else
	    *yuv422=0;
}

void adv7619_get_frameinfo(handle_t adv7619_handle,adv7619_frameinfo_t *frameinfo) 
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
   
    if (adv7619_cxt->frameinfo.is_interlace)
    {
		adv7619_cxt->frameinfo.height = adv7619_cxt->frameinfo.height;
	}
    
    if(frameinfo)
        *frameinfo=adv7619_cxt->frameinfo;
}

void adv7619_get_hdcp_flag(handle_t adv7619_handle,adv7619_frameinfo_t *frameinfo) 
{
   
}

void adv7619_set_out_format(handle_t adv7619_handle,adv7619_out_format_e format)
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
   
    io_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_03,format);
}

void adv7619_set_freerun_screen(handle_t adv7619_handle,BOOL_T is_blue_screen)
{
	adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    
    if (is_blue_screen)
    {
        cp_mask_write(adv7619_cxt,0xbf,0x01,0x01);
        printk(">>screen off\n");
	}
    else
    {
        cp_mask_write(adv7619_cxt,0xbf,0x01,0x00);
        printk(">>screen on\n");
	}
        
    if (hdmi_encrpt)
    {
		cp_mask_write(adv7619_cxt,0xbf,0x01,0x01);
		printk(">>hdcp black-screen\n");
	}    
    
}

static void adv7619_callback(adv7619_cxt_t *adv7619_cxt,adv7619_event_e event)
{
    if(adv7619_cxt)
    {
        if(adv7619_cxt->callback_func)
        {
            adv7619_cxt->callback_func(adv7619_cxt->cb_cxt,event);
        }
    }     
}

void adv7619_register_callback(handle_t adv7619_handle,adv7619_callback_t callback_func,void *cxt)
{
    adv7619_cxt_t *adv7619_cxt=adv7619_handle;
    
    if(adv7619_cxt)
    {
        if(callback_func)
        {
           adv7619_cxt->callback_func=callback_func;
           adv7619_cxt->cb_cxt=cxt;
        }
    }
    
}

void adv7619_register_callback_for_getcardid(handle_t adv7619_handle, adv7619_callback_t_for_get_cardid callback_func, void* cb_cxt_for_getid)
{
	adv7619_cxt_t *adv7619_cxt = adv7619_handle;
    
    if(adv7619_cxt)
    {
        if(callback_func)
        {
           adv7619_cxt->call_fun_get_cardid = callback_func;
           adv7619_cxt->cb_cxt_for_getid = cb_cxt_for_getid;
        }
    }
}

#if 0
static int edid_set(adv7619_cxt_t *adv7619_cxt,unsigned char *edid_buffer)
{
	int i,j;
	int ret=-1;
	U8_T check[8];
    /* set edid data for 2 segments (i=0 and i=1)*/
    for (i=0; i< /*EDID_SEG*/1; i++)
    {
		/* segment setting =0*/
		//edidctrl = rep_read(This, 0x77);
        //edidctrl = (edidctrl & ~0x10) | ((i << 4) & 0x10); /*bit4 0:seg0 1:seg1*/
        //ret = rep_write(This, 0x77, edidctrl);
        
        for (j=0; j< EDID_BYTES; j++) /* set edid data*/
        {
			 ret = edid_write(adv7619_cxt, (0x00+ i*EDID_BYTES)+j, *(edid_buffer+j)); 
		}
		sys_msleep(5); 
		if (ret ==-1)
		    printk(">>>>>>%s  fail\n",__func__);
		
	    #if 1 //check edid data
	    for (j=0; j< EDID_BYTES; j++)
	    {
		    //check[j%8] = edid_read(This, j); 
		    check[j%8] = edid_read(adv7619_cxt, 0x00+j); //
		    if (j%8 ==7)
	 	    {
			    printk(" ===========================================================\n");
	            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
	            check[0],check[1],check[2],check[3],check[4],check[5],check[6],check[7]);
		    }
	    }
	    #endif
    }
    return 0; 
}
#endif

static int adv7619_internal_edid(adv7619_cxt_t *adv7619_cxt)
{ 
	U8_T Data;
	//U8_T EdidCtrl;
    //u8 *p_edid = edid;
    U8_T edid_datas[EDID_BYTES] = { 0 };
    int ret=-1, i = 0;

	// disable HDMI port
	Data = 0xa3;
	io_write(adv7619_cxt, ADV7619_IO_REG_HDMI_LVL_INT_CLR_3, Data); 
	sys_msleep(5);
 
	Data = 0x30;
	io_write(adv7619_cxt, ADV7619_IO_REG_HPA_REG1, Data); 
	sys_msleep(5);
    /*****************************************************************/
    // disable edid
    //EdidCtrl =repeater_read(adv7619_cxt,0x74);
    repeater_mask_write(adv7619_cxt,0x74,0x01,0);
	sys_msleep(5);

	/*****************************************************************/
	// set edid data to ram
	#if 0
	if (adv7619_cxt->board_id==0x3110)
	{
	    memcpy(edid,adv7619_edid_cm311h,sizeof(edid));
	    edid_set(adv7619_cxt,edid);
	}
	#endif    

	if(g_card_id == 0)
	{
		memcpy(edid_datas, adv7619_edid_cn311h_4k, EDID_BYTES);
	}
	else if(g_card_id == 1)
	{
		memcpy(edid_datas, adv7619_edid_cn311h_1080p, EDID_BYTES);
	}

	for(i = 0; i < sizeof(edid_datas); i++)
	{
		edid_write(adv7619_cxt, i, edid_datas[i]);
	}
	
	/*****************************************************************/
	Data = 0xab;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x70, 1, &Data, 1);	// REPETER_MAP- 0x64
	ret=repeater_write(adv7619_cxt,0x70,Data);
	
	repeater_mask_write(adv7619_cxt,0x71,0x01,0);

	Data = 0x10;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x52, 1, &Data, 1);	// PORT B
	ret=repeater_write(adv7619_cxt,0x52,Data);
	
	Data = 0x00;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x53, 1, &Data, 1);
	ret=repeater_write(adv7619_cxt,0x53,Data);	

	Data = 0x10;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x72, 1, &Data, 1);	// PORT C
	ret=repeater_write(adv7619_cxt,0x72,Data);	
	
	Data = 0x00;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x73, 1, &Data, 1);
	ret=repeater_write(adv7619_cxt,0x73,Data);	

	Data = 0x10;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x74, 1, &Data, 1);	// PORT D
	ret=repeater_write(adv7619_cxt,0x74,Data);	
	
	Data = 0x00;	// change depended on edid @
	//I2CWrite(Device, REPETER_MAP, 0x75, 1, &Data, 1);
	ret=repeater_write(adv7619_cxt,0x75,Data);	
	
	/*****************************************************************/
	//enable edid
    //EdidCtrl =repeater_read(adv7619_cxt,0x74);
    repeater_mask_write(adv7619_cxt,0x74,0x01,1);
	sys_msleep(100);
	
	/*****************************************************************/
	Data = 0xa3; //HotPlug
	//I2CWrite(Device, HDMI_SLAVE, 0x6c, 1, &Data, 1);	// HDMI_SLAVE- 0x68
	ret=hdmi_write(adv7619_cxt,0x6c,Data);
	sys_msleep(5);

	Data = 0xb0;
	//I2CWrite(Device, IO_SLAVE, 0x20, 1, &Data, 1);	// IO- 0x98
	//io_write(adv7619_cxt, ADV7619_IO_REG_HPA_REG1, Data); 
	io_mask_write(adv7619_cxt, ADV7619_IO_REG_HPA_REG1,0xb0,Data);
	sys_msleep(500);
	//This is the main I2C power-down control. // CP core.//A power down control for pads of the digital output pins
	//io_mask_write(adv7619_cxt, ADV7619_IO_REG_IO_REG_0C, (ADV7619_IO_BIT_POWER_DOWN|ADV7619_IO_BIT_CP_PWRDN|ADV7619_IO_BIT_PADS_PDN),0);
   
    return 0;
}

static const U16_T chip_ids[]=
{
    0x20c1,
};

static void *adv7619_attach(i2c_model_bus_handle_t bus_handle,unsigned short i2c_addr,void *drv_set_info)
{
    adv7619_cxt_t *adv7619_cxt=NULL;
    adv7619_cfg_t *cfg=drv_set_info;
    i2c_model_handle_t i2c_mgr=NULL;
    cxt_mgr_handle_t cxt_mgr=NULL;
    task_model_handle_t task_model_handle=NULL;
    task_handle_t task_handle=NULL;
    int i;
    
    enum
    {
        NO_ERROR,
        ERROR_ALLOC_CXT,
        NO_MATCH_CHIPID,
        NO_CFG,
        ERROR_GET_TASKMGR,
        ERROR_ALLOC_TASK,
    }err=NO_ERROR;
    do
    {
        U16_T chip_id=0;
        BOOL_T match_chipid=FALSE;
        U8_T value;
        if(!cfg)
        {
            err=NO_CFG;
            break;
        }
        adv7619_cxt=mem_model_alloc_buffer(sizeof(adv7619_cxt_t));
        if(!adv7619_cxt)
        {
            err=ERROR_ALLOC_CXT;
            break;
        }
        
        adv7619_cxt->bus_handle=bus_handle;
        adv7619_cxt->addr=i2c_addr;
        chip_id=(adv7619_rd_reg(adv7619_cxt,ADV7619_IO_REG_IO_REG_RD_INFO_1)<< 8) | adv7619_rd_reg(adv7619_cxt,ADV7619_IO_REG_IO_REG_RD_INFO_2);
        for(i=0;i<sizeof(chip_ids)/sizeof(U16_T);i++)
        {
            if(chip_id==chip_ids[i])
            {
                match_chipid=TRUE;
                break;
            }
        }
        if(!match_chipid)
        {
            err=NO_MATCH_CHIPID;
            debug_msg("unknown chipid %04x\n",chip_id);
            break;
        }
     
        printk("adv7619 chipid %04x\n", chip_id); 
        
        for(i=0;i<ADV7619_MAP_COUNT;i++)
        {
           adv7619_cxt->map_addr[i]=(i2c_addr+i);
        }
        if(cfg->cp_map_addr)
            adv7619_cxt->map_addr[CP_MAP]=cfg->cp_map_addr;
        if(cfg->hdmi_map_addr)
            adv7619_cxt->map_addr[HDMI_MAP]=cfg->hdmi_map_addr;
        if(cfg->repeater_map_addr)
            adv7619_cxt->map_addr[REPEATER_MAP]=cfg->repeater_map_addr;
        if(cfg->edid_map_addr)
            adv7619_cxt->map_addr[EDID_MAP]=cfg->edid_map_addr;
        if(cfg->infoframe_map_addr)
            adv7619_cxt->map_addr[INFOFRAME_MAP]=cfg->infoframe_map_addr;
        if(cfg->cec_map_addr)
            adv7619_cxt->map_addr[CEC_MAP]=cfg->cec_map_addr;
        if(cfg->dpll_map_addr)
            adv7619_cxt->map_addr[DPLL_MAP]=cfg->dpll_map_addr;
        
        adv7619_reset(adv7619_cxt);
        sys_msleep(10);
        
        for(i=CP_MAP;i<ADV7619_MAP_COUNT;i++)
        {
            U8_T addr=adv7619_cxt->map_addr[i]<<1;
            switch(i)
            {
                case CP_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_CP_SLAVE_ADDRESS,addr);
                    break;
                case HDMI_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_HDMI_SLAVE_ADDRESS,addr);
                    break;
                case REPEATER_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_KSV_SLAVE_ADDRESS,addr);
                    break;
                case EDID_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_EDID_SLAVE_ADDRESS,addr);
                    break;
                case INFOFRAME_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_INFOFRAME_SLAVE_ADDRESS,addr);
                    break;
                case CEC_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_CEC_SLAVE_ADDRESS,addr);
                    break;
                case DPLL_MAP:
                    adv7619_wr_reg(adv7619_cxt,ADV7619_IO_REG_AFE_SLAVE_ADDRESS,addr);
                    break;
                default:
                    break;
                    
            }
            
        }
        adv7619_cxt->id=adv7619_dev_count++;
        //adv7619_cxt->board_id = subsystem_id;
        if(cfg->xtal_freq)
        {
            switch(cfg->xtal_freq)
            {
                case XTAL_FREQ_28_63636MHz:
                    io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_04,ADV7619_IO_XTAL_FREQ_SEL_MASK,ADV7619_IO_XTAL_FREQ_SEL_28_63636M);
                    break;
                case XTAL_FREQ_27MHz:
                    io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_04,ADV7619_IO_XTAL_FREQ_SEL_MASK,ADV7619_IO_XTAL_FREQ_SEL_27M);
                    break;
                case XTAL_FREQ_24_576MHz:
                    io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_04,ADV7619_IO_XTAL_FREQ_SEL_MASK,ADV7619_IO_XTAL_FREQ_SEL_24_576M);
                    break;
                case XTAL_FREQ_24MHz:
                    io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_04,ADV7619_IO_XTAL_FREQ_SEL_MASK,ADV7619_IO_XTAL_FREQ_SEL_24M);
                    break;
            }
        }
        if(cfg->out_ctrl.out_fmt)
            io_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_03,cfg->out_ctrl.out_fmt);
        
        if(cfg->out_ctrl.out_ch_order)
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_04,ADV7619_IO_OP_CH_SEL_MASK,(cfg->out_ctrl.out_ch_order) << 5);
#if 0        
        switch(cfg->out_ctrl.out_colorspace)
        {
		case ADV7619_OUT_DEFAULT:
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_ALT_GAMMA|ADV7619_IO_BIT_OP_656_RANGE,0);
            break;
        case ADV7619_OUT_YUV601:
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_ALT_GAMMA,0);
            break;
        case ADV7619_OUT_YUV709:
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_ALT_GAMMA,ADV7619_IO_BIT_ALT_GAMMA);
            break;
        case ADV7619_OUT_RGB_LIMITED:
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_OP_656_RANGE,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_OP_656_RANGE);
            break;
        case ADV7619_OUT_RGB_FULL:
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,ADV7619_IO_BIT_RGB_OUT|ADV7619_IO_BIT_OP_656_RANGE,ADV7619_IO_BIT_RGB_OUT);
            break;
        }
#else
        io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_02,0xff,0xf4);
        
#endif    
        value=0;
        if(cfg->out_ctrl.blank_data)
            value |=  ADV7619_IO_BIT_DATA_BLANK_EN;
        if(cfg->out_ctrl.insert_av_codes)
            value |=  ADV7619_IO_BIT_AVCODE_INSERT_EN;
        #if 1
        if(cfg->out_ctrl.invert_cbcr)
            value |=  ADV7619_IO_BIT_OP_SWAP_CB_CR;
        #endif
            
        io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_05,ADV7619_IO_BIT_DATA_BLANK_EN|ADV7619_IO_BIT_AVCODE_INSERT_EN|ADV7619_IO_BIT_OP_SWAP_CB_CR,value);
        if(cfg->out_ctrl.output_bus_lsb_to_msb)
            io_mask_write(adv7619_cxt,ADV7619_IO_REG_IO_REG_30,ADV7619_IO_BIT_PIXBUS_MSB_TO_LSB_REORDER,ADV7619_IO_BIT_PIXBUS_MSB_TO_LSB_REORDER);
        
            
        adv7619_core_init(adv7619_cxt);
        io_write(adv7619_cxt,ADV7619_IO_REG_INTERRUPT_MASKB_6,ADV7619_IO_BIT_CP_LOCK_CH1_MB1|ADV7619_IO_BIT_STDI_DVALID_CH1_MB1);
        i2c_mgr=i2c_model_bus_get_mgr(bus_handle);
        cxt_mgr=get_cxt_manager_from_context(i2c_mgr);
        task_model_handle=cxt_manager_get_context(cxt_mgr,TASK_MODEL_CXT_ID,0);
        if(!task_model_handle)
        {
            err=ERROR_GET_TASKMGR;
            break;
        }
        adv7619_cxt->taskmgr=task_model_handle;
        cxt_manager_ref_context(adv7619_cxt->taskmgr);
        task_handle=task_model_create_task(task_model_handle,adv7619_task,adv7619_cxt,"adv7619_task");
        if(!task_handle)
        {
            err=ERROR_ALLOC_TASK;
            break;
        }
        //free run screen color
        cp_mask_write(adv7619_cxt,0xc0,0xff,16);
        cp_mask_write(adv7619_cxt,0xc1,0xff,128);
        cp_mask_write(adv7619_cxt,0xc2,0xff,128);
        cp_write(adv7619_cxt, 0xbf, 0x05);
        adv7619_internal_edid(adv7619_cxt);
       
        #if 0
	    io_write(adv7619_cxt, 0x6c, 0xa3); 
	    sys_msleep(5);
	    io_write(adv7619_cxt, 0x20, 0xb0); 
        sys_msleep(5);
        #endif
        
        adv7619_cxt->lock = ADV7619_UNLOCK;
        adv7619_cxt->detected_audiorate =ADV7619_AUDIO_RATE_96000; 
        
        adv7619_cxt->task=task_handle;
        task_model_run_task_period(task_model_handle,task_handle,600,-1,NULL); 
        
        
    }while(0);
    
    
    
    if(err!=NO_ERROR)
    {
        debug_msg("%s err %d\n",__func__,err);
        switch(err)
        {
                mem_model_free_buffer(adv7619_cxt);
        case ERROR_ALLOC_TASK:
        case ERROR_GET_TASKMGR:
        case ERROR_ALLOC_CXT:
        case NO_CFG:
             break;   
        default:
             break;
        }
    }
    
    return adv7619_cxt;
}

static void adv7619_detach(void *cxt)
{
    adv7619_cxt_t *adv7619_cxt=cxt;
    if(adv7619_cxt)
    {
        int i;
        if(adv7619_cxt->task)
        {
            task_model_release_task(adv7619_cxt->taskmgr,adv7619_cxt->task);
            cxt_manager_unref_context(adv7619_cxt->taskmgr);
        }    
        for(i=0;i<sizeof(adv7619_cxt->attr_handles)/sizeof(void *);i++)
            trace_attr_free(adv7619_cxt->attr_handles[i]);
        mem_model_free_buffer(adv7619_cxt);
    }
}
I2C_MODEL_DRV_DECLARE(ADV7619_ID);

int adv7619_get_video_time_list(handle_t cxt, U8_T* val, int* size)
{
	int i = 0;
	adv7619_cxt_t *adv7619_cxt = cxt;

	if((adv7619_cxt == NULL) || (*size < EDID_BYTES))
	{
		printk("adv7619_cxt = null, size = null!\n");
		return -1;
	}

	for(i = 0; i < EDID_BYTES; i++)
	{
		*(val + i) = edid_read(adv7619_cxt, 0 + i);
	}

	*size = EDID_BYTES;

	return 0;
}

int adv7619_set_video_time_list(handle_t cxt, U8_T* edid_data, int size)
{
	int i = 0;
	U8_T val[EDID_BYTES] = { 0 };
	U8_T check[8] = { 0 };
	adv7619_cxt_t *adv7619_cxt = (adv7619_cxt_t *)cxt;
	if((adv7619_cxt == NULL))
	{
		printk("[error]:adv7619_cxt = null or val = null!\n");
		return -1;
	}

	for(i = 0; i < size; i++)
	{
		edid_write(adv7619_cxt, i, edid_data[i]);
	}

	printk(" check write===========================================================\n");
	for (i=0; i < 255; i++)
    {
	    *(val + i) = edid_read(adv7619_cxt, 0x00+i);
	    check[i%8] = *(val + i);
	    if (i%8 ==7)
 	    {
		    printk(" check write===========================================================\n");
            printk(" 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x\n",
            check[0],check[1],check[2],check[3],check[4],check[5],check[6],check[7]);
	    }
    }

    adv7619_callback(adv7619_cxt, ADV7619_HOTPLUG);

    return 0;
}

