 /*
 *Driver for the card pcie bridge.

 *Copyright (C) 2019 AVerMedia TECHNOLOGIES, Inc.

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/kernel.h>
#include <linux/time.h>
#include <linux/ktime.h>
#include "linux/delay.h"
#include "linux/device.h"
#include "linux/slab.h"

#include "sys.h"

typedef struct
{
    sys_spin_lock_t interface;
    sys_spin_lock_level_e level;
    spinlock_t lock;
    unsigned long flags;
}sys_spin_lock_cxt_t;

typedef struct
{
    atomic_t value;
}sys_atomic_cxt_t;

void sys_msleep(unsigned ms)
{
    msleep(ms);
}

int sys_sprintf(char *buf, const char * fmt, ...)
{
    va_list args;
    int i;

    va_start(args, fmt);
    i = vsprintf(buf, fmt, args);
    va_end(args);
    return i;
    
}

int sys_sscanf(const char *buf, const char *fmt, ...)
{
       va_list args;
       int i;
 
       va_start(args, fmt);
       i = vsscanf(buf, fmt, args);
       va_end(args);

       return i;
}

unsigned long long sys_gettimestamp()
{
    struct timespec ts;
    unsigned long long timestamp;
    
    ktime_get_ts(&ts);
    timestamp=ts.tv_sec ;
    timestamp*=1000000000;
    timestamp+=ts.tv_nsec;
    
    return timestamp;
}

void *sys_get_drvdata(device_handle_t device_handle)
{
    struct device *dev=device_handle;
    if(dev)
	return dev_get_drvdata(dev);
    return NULL;    
}

sys_spin_lock_t *sys_new_spinlock(const char *name,sys_spin_lock_level_e level)
{
    sys_spin_lock_t *lock=NULL;
    sys_spin_lock_cxt_t *lock_cxt=kzalloc(sizeof(sys_spin_lock_cxt_t),GFP_KERNEL);
    
    if(lock_cxt)
    {
        lock_cxt->interface.name=name;
        lock_cxt->level=level;
        spin_lock_init(&lock_cxt->lock);
        lock=&lock_cxt->interface;
    }
    
    return lock;
}

void sys_spin_lock(sys_spin_lock_t *lock)
{
    sys_spin_lock_cxt_t *lock_cxt=container_of(lock,sys_spin_lock_cxt_t,interface);
    
    if(lock_cxt)
    {
        switch(lock_cxt->level)
        {
        case SYS_SPINLOCK_LEVEL_PROC:
            spin_lock(&lock_cxt->lock);
            break;
        case SYS_SPINLOCK_LEVEL_BH:
            spin_lock_bh(&lock_cxt->lock);
            break;
        case SYS_SPINLOCK_LEVEL_IRQ:
            spin_lock_irqsave(&lock_cxt->lock,lock_cxt->flags);
            break;
        default:
            break;
        }   
    }
    
}

void sys_spin_unlock(sys_spin_lock_t *lock)
{
    sys_spin_lock_cxt_t *lock_cxt=container_of(lock,sys_spin_lock_cxt_t,interface);
    
    if(lock_cxt)
    {
        switch(lock_cxt->level)
        {
        case SYS_SPINLOCK_LEVEL_PROC:
            spin_unlock(&lock_cxt->lock);
            break;
        case SYS_SPINLOCK_LEVEL_BH:
            spin_unlock_bh(&lock_cxt->lock);
            break;
        case SYS_SPINLOCK_LEVEL_IRQ:
            spin_unlock_irqrestore(&lock_cxt->lock,lock_cxt->flags);
            break;
        default:
            break;
        }    
    }
}

void sys_release_spinlock(sys_spin_lock_t *lock)
{
    sys_spin_lock_cxt_t *lock_cxt=container_of(lock,sys_spin_lock_cxt_t,interface);
    
    if(lock_cxt)
    {
        kfree(lock_cxt);
    }
    
}

sys_atomic_t sys_new_atomic()
{
    sys_atomic_cxt_t *sys_atomic_cxt=kzalloc(sizeof(sys_spin_lock_cxt_t),GFP_KERNEL);
    if(sys_atomic_cxt)
    {
        
    }
    return sys_atomic_cxt;
    
}

void sys_release_atomic(sys_atomic_t atomic)
{
    sys_atomic_cxt_t *sys_atomic_cxt=atomic;
    if(sys_atomic_cxt)
    {
        kfree(sys_atomic_cxt);
    }
    
}

void sys_atomic_set(sys_atomic_t atomic,int value)
{
    sys_atomic_cxt_t *sys_atomic_cxt=atomic;
    if(sys_atomic_cxt)
    {
        atomic_set(&sys_atomic_cxt->value,value);
    }
}

int  sys_atomic_get(sys_atomic_t atomic)
{
    sys_atomic_cxt_t *sys_atomic_cxt=atomic;
    int value=0;
    if(sys_atomic_cxt)
    {
        value=atomic_read(&sys_atomic_cxt->value);
    }
    return value;
}


    

